from appyter.cli import cli
from click_default_group import DefaultGroup

@cli.group(cls=DefaultGroup, default='nbtemplate-json', help='Serialize various formats from appyter inspection')
def nbinspect():
  pass



import re

from appyter.extras.catalog_integration.user_config import get_user_config

async def prepare_storage(data):
  ''' Safely translate public facing storage URIs into internally operable fsspec URIs
  '''
  if data.get('_storage'):
    # cavatica storage provider, obtain CAVATICA API Key and instantiate sbfs
    m = re.match(r'^cavatica://(?P<user>[a-z0-9_-]+)/(?P<project>[a-z0-9_-]+)$', data['_storage'])
    if m:
      user_config = await get_user_config(data.get('_auth'), data.get('_config'))
      if not user_config.get('cavatica_api_key'): raise PermissionError
      assert re.match(r'^[0-9a-f]+$', user_config['cavatica_api_key']) is not None, 'Malformed CAVATICA API Key'
      return f"writecache::chroot::sbfs://{m.group('user')}/{m.group('project')}/appyter/#?sbfs.auth_token={user_config['cavatica_api_key']}"
    else:
      raise Exception('Unrecognized storage URI')



import socketio
import logging
logger = logging.getLogger(__name__)

from appyter.ext.socketio.priority_queued_emit import PriorityQueuedEmitMixin
from appyter.ext.socketio.chunked_emit import ChunkedEmitMixin

class AsyncServer(
  ChunkedEmitMixin,
  PriorityQueuedEmitMixin,
  socketio.AsyncServer,
):
  ''' Server with packet forwarding.
  `async with` is necessary for initialization
  '''
  def __init__(self, *args, **kwargs):
    ChunkedEmitMixin.__init__(self)
    PriorityQueuedEmitMixin.__init__(self)
    socketio.AsyncServer.__init__(self, *args, **kwargs)
    self._listeners = { 'forward': [] }

  async def __aenter__(self):
    await super().__aenter__()
    super().on('forward')(self.forward)
    self._emit_enabled.set()
    return self
  
  async def __aexit__(self, *args):
    await super().__aexit__(*args)

  async def forward(self, sid, data):
    logger.debug(f"{sid} forward {data['event']} to {data['to']}")
    await super().emit(data['event'], data['data'], priority=max(0, data.get('priority', 0)) + 1, to=data['to'], skip_sid=sid)
    for listener in self._listeners['forward']:
      await listener(sid, data)

  def on(self, event):
    if event == 'forward':
      return self._listeners['forward'].append
    else:
      return super().on(event)



import contextlib
import typing as t

@contextlib.asynccontextmanager
async def adhoc_ssl_context(
  cn: t.Optional[str] = None,
  protocol: t.Optional[int] = None,
):
  '''
  From https://github.com/pallets/werkzeug/blob/main/src/werkzeug/serving.py
  Produces an adhoc SSL cert and returns an SSL context for https in development.
  '''
  import os
  import ssl
  import tempfile
  from datetime import datetime as dt
  from datetime import timedelta, timezone

  from cryptography import x509
  from cryptography.hazmat.backends import default_backend
  from cryptography.hazmat.primitives import hashes, serialization
  from cryptography.hazmat.primitives.asymmetric import rsa
  from cryptography.x509.oid import NameOID
  backend = default_backend()
  pkey = rsa.generate_private_key(
    public_exponent=65537, key_size=2048, backend=backend
  )
  if cn is None: cn = '*'
  subject = x509.Name([
    x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Dummy Certificate"),
    x509.NameAttribute(NameOID.COMMON_NAME, cn),
  ])
  backend = default_backend()
  cert = (
      x509.CertificateBuilder()
      .subject_name(subject)
      .issuer_name(subject)
      .public_key(pkey.public_key())
      .serial_number(x509.random_serial_number())
      .not_valid_before(dt.now(timezone.utc))
      .not_valid_after(dt.now(timezone.utc) + timedelta(days=365))
      .add_extension(x509.ExtendedKeyUsage([x509.OID_SERVER_AUTH]), critical=False)
      .add_extension(x509.SubjectAlternativeName([x509.DNSName(cn)]), critical=False)
      .sign(pkey, hashes.SHA256(), backend)
  )
  cert_handle, cert_file = tempfile.mkstemp()
  pkey_handle, pkey_file = tempfile.mkstemp()
  os.write(cert_handle, cert.public_bytes(serialization.Encoding.PEM))
  os.write(
      pkey_handle,
      pkey.private_bytes(
          encoding=serialization.Encoding.PEM,
          format=serialization.PrivateFormat.TraditionalOpenSSL,
          encryption_algorithm=serialization.NoEncryption(),
      ),
  )

  os.close(cert_handle)
  os.close(pkey_handle)

  if protocol is None:
      protocol = ssl.PROTOCOL_TLS_SERVER
  
  ctx = ssl.SSLContext(protocol)
  ctx.load_cert_chain(cert_file, pkey_file)

  try:
    yield ctx
  finally:
    os.remove(pkey_file)
    os.remove(cert_file)



import os
import traceback
import logging
from appyter.ext.json import try_json_loads
logger = logging.getLogger(__name__)

from appyter.execspec.spec import AbstractExecutor

class LocalExecutor(AbstractExecutor):
  ''' Run executions with the same process
  
  Not ideal in most circumstances:
  - it can leak errors to the notebook
  - internal state can get messed with
  '''
  protocol = 'local'

  def __init__(self, url=None, **kwargs) -> None:
    super().__init__(url=url, **kwargs)

  async def _submit(self, emit=None, **job):
    from appyter.render.nbexecute import nbexecute_async
    try:
      fuse = job.get('fuse')
      if fuse is None: fuse = try_json_loads(os.environ.get('APPYTER_FUSE'))
      if fuse is None: fuse = not job.get('debug', False)
      await nbexecute_async(
        cwd=f"storage://{job['cwd']}",
        ipynb=job['ipynb'],
        emit=emit,
        fuse=fuse,
      )
    except:
      logger.error(traceback.format_exc())
      await emit({ 'type': 'error', 'data': 'An error occurred while initializing the execution...' })
    finally:
      await emit(None)

  async def _run(self, **job):
    import asyncio
    from appyter.ext.fsspec.storage import ensure_storage
    async with ensure_storage(str(job.get('storage', 'file://'))):
      msg_queue = asyncio.Queue()
      task = asyncio.create_task(self._submit(emit=msg_queue.put, **job))
      while True:
        msg = await msg_queue.get()
        if msg: yield msg
        msg_queue.task_done()
        if not msg: break
      await task



import logging
logger = logging.getLogger(__name__)

from appyter.execspec.spec import AbstractExecutor
from appyter.ext.asyncio.try_n_times import async_try_n_times
from appyter.ext.asyncio.helpers import ensure_async
from appyter.ext.dict import dict_merge

def endless_watch(*args, **kwargs):
  from kubernetes import watch
  w = watch.Watch()
  s = iter(w.stream(*args, **kwargs))
  v = None
  while True:
    try:
      event = next(s)
      v = event['object'].metadata.resource_version
      yield event
    except StopIteration:
      s = iter(w.stream(*args, **kwargs, resource_version=v))

class KubernetesExecutor(AbstractExecutor):
  ''' Run executions as kubernetes jobs
  Example uri:
  kube::maayanlab/myimage:latest?namespace=default
  '''
  protocol = 'kube'

  def __init__(self, url=None, **kwargs) -> None:
    super().__init__(url=url, **kwargs)
    from kubernetes import config
    config.load_incluster_config()

  def _submit(self, job, namespace='default'):
    from kubernetes import client
    batchV1 = client.BatchV1Api()
    batchV1.create_namespaced_job(
      namespace=namespace,
      body=client.V1Job(
        api_version='batch/v1',
        kind='Job',
        metadata=client.V1ObjectMeta(
          name=job['id'],
          annotations={
            f"container.apparmor.security.beta.kubernetes.io/appyter-{job['id']}": 'unconfined'
          },
        ),
        spec=client.V1JobSpec(
          template=client.V1PodTemplateSpec(
            spec=client.V1PodSpec(
              restart_policy='Never',
              containers=[
                client.V1Container(
                  name=f"appyter-{job['id']}",
                  image=self.url,
                  command=[
                    'appyter', 'nbexecute',
                    *(
                      f"--{k}={v}" if len(k) > 1 else f"-{k}{v}"
                      for k, v in dict_merge(
                        {
                          's': job['url'],
                          'w': f"storage://{job['cwd']}",
                          'data-dir': job['storage'],
                          'fuse': 'true',
                        },
                        **self.executor_options.get('args', {})
                      ).items()
                    ),
                    job['ipynb'],
                  ],
                  env=[
                    client.V1EnvVar(
                      name=key,
                      value=value,
                    )
                    for key, value in self.executor_options.get('env', {}).items()
                  ],
                  security_context=client.V1SecurityContext(
                    privileged=True,
                    capabilities=client.V1Capabilities(
                      add=['SYS_ADMIN'],
                    ),
                  ),
                  volume_mounts=[
                    client.V1VolumeMount(
                      name='fuse',
                      mount_path='/dev/fuse',
                    )
                  ],
                  resources=client.V1ResourceRequirements(
                    **self.executor_options.get('resources', {})
                  ),
                ),
              ],
              volumes=[
                client.V1Volume(
                  name='fuse',
                  host_path=client.V1HostPathVolumeSource(
                    path='/dev/fuse',
                  )
                ),
              ],
            ),
          ),
          ttl_seconds_after_finished=120,
          backoff_limit=1,
        ),
      ),
    )
  submit = ensure_async(_submit)

  def _wait_for(self, run_id, namespace='default', debug=False):
    from kubernetes import client
    coreV1 = client.CoreV1Api()
    batchV1 = client.BatchV1Api()
    for event in endless_watch(batchV1.list_namespaced_job, namespace, 
      label_selector=f"job-name={run_id}"
    ):
      logger.debug(str(event))
      event_type = event['type']
      event_job = event['object']
      if event_type == 'MODIFIED':
        if event_job.status.succeeded or event_job.status.failed:
          break
    logger.info(f"{run_id} completed")
    return 0
  wait_for = ensure_async(_wait_for)

  async def _run(self, **job):
    yield dict(type='status', data=f"Submitting appyter for execution..")
    await async_try_n_times(3, self._submit, job, namespace=self.executor_options.get('namespace', 'default'))
    yield dict(type='status', data=f"Queued successfully, your execution will begin when resources are available..")
    # TODO: kubernetes probably reports some additional status we can report
    await self._wait_for(job['id'], namespace=self.executor_options.get('namespace', 'default'), debug=self.executor_options.get('debug'))
    yield dict(type='status', data=f"Job completed.")



import asyncio
import logging
logger = logging.getLogger(__name__)

async def stream_readline_to_queue(stream, queue, chunk_size=65536):
  ''' Safely yield lines without running into buffer limit
  '''
  try:
    linebuf = b''
    while True:
      buf = await stream.read(chunk_size)
      if not buf: break
      linebuf += buf
      while True:
        msg, nl, rest = linebuf.partition(b'\n')
        if nl:
          await queue.put((msg + nl, False))
          linebuf = rest
        else:
          break
    if linebuf: await queue.put((linebuf, False))
  except Exception as e:
    await queue.put((e, True))
  else:
    await queue.put((None, True))

async def ensure_terminated(proc: asyncio.subprocess.Process):
  try:
    if proc.returncode is None:
      proc.terminate()
    await asyncio.sleep(5000)
  finally:
    if proc.returncode is None:
      proc.kill()

async def sh(*args, chunk_size=65536, **kwargs):
  ''' An opinionated asyncio shell that logs stderr & yields stdout, done
  Unlike standard create_subprocess_exec: LimitOverrunError should not be possible.
  '''
  stdout_queue = asyncio.Queue()
  logger.debug(' '.join(repr(arg) if ' ' in arg else arg for arg in args))
  proc = await asyncio.create_subprocess_exec(
    *args,
    stdout=asyncio.subprocess.PIPE,
    limit=chunk_size * 2,
    **kwargs,
  )
  try:
    reader = asyncio.create_task(stream_readline_to_queue(proc.stdout, stdout_queue, chunk_size=chunk_size))
    while True:
      msg, done = await stdout_queue.get()
      if isinstance(msg, bytes):
        yield msg, False
      elif isinstance(msg, Exception):
        raise msg
      stdout_queue.task_done()
      if done: break
    await reader
    yield await proc.wait(), True
  except:
    raise
  finally:
    asyncio.create_task(ensure_terminated(proc))



import logging
logger = logging.getLogger(__name__)

class ForwardingMixin:
  ''' A mixin for converting `to=` messages on the client to `forward` events to be handled by the server
  '''
  async def emit(self, evt, data, to=None, priority=0, **kwargs):
    if to is not None:
      logger.debug(f"Message forwarding {evt} to {to}")
      return await super().emit('forward', dict(event=evt, data=data, to=to, priority=priority))
    else:
      return await super().emit(evt, data, priority=priority)



async def alist(agen):
  ''' Resolve all values from an asynchronous generator like it was an iterator
  '''
  L = []
  async for el in agen:
    L.append(el)
  return L

def ensure_list(L):
  if type(L) == list:
    return L
  else:
    return [L]

def collapse(L):
  if len(L) == 1:
    return L[0]
  else:
    return L

def one_and_only(it):
  it = iter(it)
  try:
    ret = next(it)
    try:
      next(it)
      raise Exception('Expected one got multiple')
    except StopIteration:
      pass
  except StopIteration:
    raise Exception('Expected one got none')
  return ret



import fsspec
from appyter.ext.fsspec.chroot import ChrootFileSystem

_singletons = {}

class SingletonFileSystem(ChrootFileSystem):
  ''' Instantiation of a singleton filesystem will register it by protocol after which
  the protocol can be used to re-instantiate the singleton instance. Serialization
  works for this enabling the singleton to re-register itself across process boundaries.
  '''
  root_marker = ''

  def __init__(self, proto=None, fs=None, **kwargs):
    global _singletons
    if proto is None:
      if self.__class__.protocol == 'chroot':
        raise ValueError('proto is required')
      proto = self.__class__.protocol
    if fs is None:
      fs = _singletons[proto]
    super(SingletonFileSystem, self).__init__(fs=fs, **kwargs)
    if proto not in _singletons:
      _singletons[proto] = self
      class SingletonFileSystem_(SingletonFileSystem):
        protocol = proto
        def to_json(self):
          global _singletons
          return _singletons[proto].to_json()
      SingletonFileSystem_.__name__ = 'SingletonFileSystem'
      fsspec.register_implementation(proto, SingletonFileSystem_)
    self.protocol = proto
    self.stack_count = 0
  
  def __enter__(self):
    if not self.stack_count:
      super(SingletonFileSystem, self).__enter__()
    self.stack_count += 1
    return self

  def __exit__(self, type, value, traceback):
    self.stack_count -= 1
    if not self.stack_count:
      super(SingletonFileSystem, self).__exit__(type, value, traceback)



def main():
  import os
  import appyter.ext.multiprocessing
  # load dotenv if present
  os.environ['PYTHONUTF8'] = '1'
  from dotenv import load_dotenv
  load_dotenv()
  # load everything permitting injection to cli
  from pathlib import Path
  from appyter.ext.importlib import importdir
  importdir(Path(__file__).parent, __package__, globals())
  # grab cli
  from appyter.cli import cli
  cli()

if __name__ == '__main__':
  main()



from appyter.ext.urllib import join_url, parent_url

async def get_user_config(auth=None, config=None):
  if not auth: raise PermissionError
  import aiohttp
  async with aiohttp.ClientSession(
    headers={
      'Authorization': f"Bearer {auth}",
    },
    raise_for_status=True,
  ) as session:
    async with session.post(
      join_url(
        parent_url(config['PUBLIC_URL']),
        'postgrest/rpc/user_config',
      ),
      json=dict(config=None),
    ) as res:
      return await res.json()



import asyncio
import itertools
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.subprocess import sh

async def _test_subprocess(expectation, chunk_size=None):
  i = iter(itertools.count())
  async for msg, done in sh('cat', __file__, chunk_size=chunk_size):
    assert_eq(msg, expectation[next(i)])

def test_subprocess():
  with open(__file__, 'rb') as fr:
    expectation = fr.readlines()
  expectation.append(0)
  # ensure it works with lines bigger than chunk_size
  asyncio.new_event_loop().run_until_complete(_test_subprocess(expectation, chunk_size=5))
  # ensure it works with chunks bigger than the lines
  asyncio.new_event_loop().run_until_complete(_test_subprocess(expectation, chunk_size=len(expectation)*2))



import click
import pathlib

from appyter.cli import cli
from appyter.ext.click import click_argument_setenv

def ipynb_options_from_sys_argv(func):
  import sys
  if len(sys.argv) >= 3 and sys.argv[1] == 'commandify' and sys.argv[2][0] != '-':
    import os
    import fsspec
    import functools
    from appyter.context import get_env, get_jinja2_env
    from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
    from appyter.ext.dict import dict_filter_none
    from appyter.parse.nb import nb_from_ipynb_io
    from appyter.ext.asyncio.event_loop import with_event_loop
    #
    arg = pathlib.Path(sys.argv[2])
    assert arg.exists()
    ipynb = arg.name
    cwd = arg.parent.absolute()
    #
    with with_event_loop():
      with fsspec.open(str(cwd / ipynb), 'r') as fr:
        nbtemplate = nb_from_ipynb_io(fr)
    #
    env = get_jinja2_env(
      config=get_env(cwd=str(cwd), ipynb=ipynb, mode='inspect', safe_mode=False),
    )
    fields = list(dict_filter_none({
      field.args['name']: field.to_click()
      for field in parse_fields_from_nbtemplate(env, nbtemplate, deep=True)
    }).values())
    for field in fields:
      func = click.option(*field[0], **field[1])(func)
    #
    @functools.wraps(func)
    def wrapper(**kwargs):
      with with_event_loop():
        return func(dict(cwd=str(cwd), ipynb=ipynb, nbtemplate=nbtemplate), **kwargs)
    return wrapper
  return func

@cli.group(help='Construct an operable command-line from an appyter')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def commandify(ipynb=None):
  pass

@commandify.command(help='Construct a notebook from an appyter and its arguments')
@click.option('-o', type=str, metavar='FILE', default='-', help='Output notebook')
@ipynb_options_from_sys_argv
def nbconstruct(ctx, o=None, **kwargs):
  if o == '-': o = '/dev/stdout'
  import fsspec
  from appyter.context import get_env, get_jinja2_env
  from appyter.render.nbconstruct import render_nb_from_nbtemplate
  from appyter.parse.nb import nb_to_ipynb_io
  env = get_jinja2_env(
    config=get_env(cwd=ctx['cwd'], ipynb=ctx['ipynb'], mode='construct'),
    context=kwargs,
  )
  nb = render_nb_from_nbtemplate(env, ctx['nbtemplate'], data=kwargs)
  with fsspec.open(o, 'w') as fw:
    nb_to_ipynb_io(nb, fw)

@commandify.command(help='Construct and execute an appyter')
@click.option('-s', type=str, metavar='URI', default='file:///dev/stderr', help='Status stream')
@click.option('-e', type=str, metavar='URI', default='local', help='Executor')
@click.option('-o', type=str, metavar='FILE', default=None, help='Output notebook')
@click.option('-w', type=str, metavar='DIR', default=None, help='Working directory')
@ipynb_options_from_sys_argv
def run(ctx, s=None, e=None, o=None, w=None, **kwargs):
  if s == '-':
    s = 'file:///dev/stderr'
  #
  if o is None:
    if w is None:
      o = 'file:///dev/stdout'
  elif o == '-':
    o = 'file:///dev/stdout'
  #
  import fsspec
  import shutil
  from appyter.context import get_env, get_jinja2_env
  from appyter.parse.nb import nb_to_ipynb_io
  from appyter.render.nbconstruct import render_nb_from_nbtemplate
  from appyter.ext.emitter import url_to_emitter
  from appyter.execspec.core import url_to_executor
  from appyter.ext.tempfile import tempdir
  from appyter.ext.asyncio.helpers import ensure_sync
  # render notebook
  env = get_jinja2_env(
    config=get_env(cwd=ctx['cwd'], ipynb=ctx['ipynb'], mode='construct', safe_mode=False),
    context=kwargs,
  )
  nb = render_nb_from_nbtemplate(env, ctx['nbtemplate'], data=kwargs)
  # write into temporary directory
  with tempdir(w) as tmp_dir:
    with (tmp_dir/ctx['ipynb']).open('w') as fw:
      nb_to_ipynb_io(nb, fw)
    # execute notebook, sending status updates to `s`
    with ensure_sync(url_to_emitter(s)) as emitter:
      emitter = ensure_sync(emitter)
      with url_to_executor(e) as executor:
        for msg in executor.run(
          cwd=str(tmp_dir),
          ipynb=ctx['ipynb'],
          fuse=False,
        ):
          emitter(msg)
    # write output notebook, to `o`
    if o:
      with (tmp_dir/ctx['ipynb']).open('r') as fr:
        with fsspec.open(o, 'w') as fw:
          shutil.copyfileobj(fr, fw)



'''
WARNING: This blueprint may open an XSS vulnerability, and is thus opt-in through the storage-file-field extra, use with caution.
'''
import logging
logger = logging.getLogger(__name__)

def StorageFileField(flask_app, url_prefix=None):
  if 'storage-file-field' in flask_app.config['EXTRAS']:
    logger.info("Enabling storage file field...")

    from flask import Blueprint, abort, jsonify, request
    import traceback
    from appyter.ext.fsspec.core import url_to_chroot_fs
    from appyter.ext.flask import route_join_with_or_without_slash

    blueprint = Blueprint('StorageFileField', __name__)

    @route_join_with_or_without_slash(blueprint, '', methods=['GET'])
    def index():
      abort(404)

    @route_join_with_or_without_slash(blueprint, 'ls', methods=['GET'])
    @route_join_with_or_without_slash(blueprint, 'ls', '<path:path>', methods=['GET'])
    def ls(path=''):
      try:
        fs = url_to_chroot_fs(path + '#?' + request.query_string.decode())
        return jsonify(fs.ls('/'))
      except KeyboardInterrupt:
        raise
      except Exception:
        logger.error(traceback.format_exc())
        abort(500)

    @route_join_with_or_without_slash(blueprint, 'info', methods=['GET'])
    @route_join_with_or_without_slash(blueprint, 'info', '<path:path>', methods=['GET'])
    def info(path=''):
      try:
        fs = url_to_chroot_fs(path + '#?' + request.query_string.decode())
        return jsonify(fs.info('/'))
      except KeyboardInterrupt:
        raise
      except Exception:
        logger.error(traceback.format_exc())
        abort(500)

    # @route_join_with_or_without_slash(blueprint, 'cat', methods=['GET'])
    # @route_join_with_or_without_slash(blueprint, 'cat', '<path:path>', methods=['GET'])
    # def cat(path=''):
    #   try:
    #     fs, fs_path = url_to_fs_ex(path + '#?' + request.query_string.decode())
    #     return fs.cat(fs_path)
    #   except KeyboardInterrupt:
    #     raise
    #   except Exception:
    #     logger.error(traceback.format_exc())
    #     abort(500)

    # @route_join_with_or_without_slash(blueprint, 'read_block', methods=['GET'])
    # @route_join_with_or_without_slash(blueprint, 'read_block', '<path:path>', methods=['GET'])
    # def read_block(path=''):
    #   try:
    #     m = re.match(r'^(bytes )?(\d+)-(\d+)$', request.headers['Range'])
    #     _, start, end = m.groups()
    #     fs, fs_path = url_to_fs_ex(path + '#?' + request.query_string.decode())
    #     return fs.read_block(fs_path, start, end - start)
    #   except KeyboardInterrupt:
    #     raise
    #   except Exception:
    #     logger.error(traceback.format_exc())
    #     abort(500)

    flask_app.register_blueprint(blueprint, url_prefix=url_prefix)



from appyter.execspec.spec import AbstractExecutor
from appyter.ext.asyncio.try_n_times import async_try_n_times
from appyter.ext.dict import dict_merge

class DispatchExecutor(AbstractExecutor):
  ''' Submit executions to be run by an appyter orchestration dispatcher

  This executor only supports submit and not wait_for/run
   as it submits jobs to a queue which manages the execution.
  
  usage:
  dispatch::http://appyter-orchestrator:5000?params.executor=docker::maayanlab/appyter-example:latest
  '''
  protocol = 'dispatch'

  async def __aenter__(self):
    import aiohttp
    self.session = aiohttp.ClientSession(
      headers=dict(
        {'Content-Type': 'application/json'},
        **self.executor_options.get('headers', {}),
      )
    )
    self.client = await self.session.__aenter__()
    return await super().__aenter__()
  
  async def __aexit__(self, type, value, traceback):
    await self.session.__aexit__(type, value, traceback)
    await super().__aexit__(type, value, traceback)

  async def _submit(self, job):
    async with self.client.post(
      self.url,
      json=dict_merge(
        self.executor_options.get('params',{}),
        **dict(job, storage=job['storage']),
      ),
    ) as resp:
      queue_size = await resp.json()
      return queue_size

  async def _run(self, **job):
    yield dict(type='status', data=f"Submitting appyter for execution..")
    queue_size = await async_try_n_times(3, self._submit, job)
    yield dict(type='status', data=f"Queued successfully, you are at position {queue_size}, your execution will begin when resources are available..")
    # TODO: room drop hotfix until we have queue tracking
    import asyncio
    await asyncio.sleep(60)



import uuid
import logging
logger = logging.getLogger(__name__)

from appyter.ext.socketio import AsyncServer

socketio = AsyncServer(async_mode='aiohttp', cors_allowed_origins=[])

@socketio.on('connect')
async def _(sid, environ):
  request = environ['aiohttp.request']
  uid = request.headers.get('Authorization')
  if uid is None:
    uid = str(uuid.uuid4())
  logger.debug(f"connect: {uid} ({sid})")
  async with socketio.session(sid) as sess:
    sess['uid'] = uid
    sess['request_url'] = request_url = f"{request.scheme}://{request.host}{request.path}"
    if not request.app['config']['DEBUG']:
      public_url = request.app['config'].get('PUBLIC_URL')
      request_url = sess['request_url']
      if not public_url or not request_url.startswith(public_url):
        logger.warning(f"This could cause issues in production:\n{request_url=} {public_url=}")
    sess['config'] = request.app['config']
    sess['executor'] = request.app['executor']
  #
  from appyter.render.flask_app.room_manager import ensure_joined
  await ensure_joined(sid)



import logging

from appyter.render.flask_app.prepare import prepare_request
logger = logging.getLogger(__name__)

def locate(flask_app, url_prefix=None):
  if 'catalog-integration' in flask_app.config['EXTRAS']:
    logger.info("Enabling locate blueprint...")

    from flask import Blueprint, abort, jsonify, request
    import traceback
    from appyter.ext.flask import route_join_with_or_without_slash
    from appyter.extras.catalog_integration.userfs import create_userfs
    from appyter.ext.asyncio.helpers import ensure_sync

    blueprint = Blueprint('locate', __name__)

    @route_join_with_or_without_slash(blueprint, '', methods=['GET'])
    def index():
      abort(404)

    @route_join_with_or_without_slash(blueprint, 'ls', methods=['GET'])
    @route_join_with_or_without_slash(blueprint, 'ls', '<path:path>', methods=['GET'])
    def ls(path=''):
      if not path.startswith('user://'):
        return abort(404)
      try:
        data = prepare_request(request)
        user_fs = ensure_sync(create_userfs(data))
      except:
        logger.error(traceback.format_exc())
        return abort(401)
      try:
        with user_fs as user_fs:
          return jsonify(user_fs.ls(path[len('user://'):], detail=True))
      except KeyboardInterrupt:
        raise
      except Exception:
        logger.error(traceback.format_exc())
        return abort(500)

    flask_app.register_blueprint(blueprint, url_prefix=url_prefix)



import asyncio
import itertools
import logging
logger = logging.getLogger(__name__)

class PriorityQueuedEmitMixin:
  ''' A mixin for queuing `emit` calls to get triggered sequentially when `emit_enabled` event is set
  '''
  async def __aenter__(self):
    self._emit_enabled = asyncio.Event()
    self._emit_counter = iter(itertools.count())
    self._emit_queue = asyncio.PriorityQueue()
    self._emit_dispatcher_task = asyncio.create_task(self._emit_dispatcher())

  async def __aexit__(self, *args):
    if self._emit_queue.qsize() != 0:
      logger.warning(f"{self._emit_queue.qsize()} items in queue weren't processed...")
    try:
      self._emit_dispatcher_task.cancel()
      await self._emit_dispatcher_task
    except asyncio.CancelledError:
      pass

  async def _emit_dispatcher(self):
    while True:
      _, _, args, kwargs = await self._emit_queue.get()
      try:
        await self._emit_enabled.wait()
        await super().emit(*args, **{k:v for k,v in kwargs.items() if v})
      except asyncio.CancelledError:
        raise
      except:
        import traceback
        logger.error(traceback.format_exc())
      finally:
        self._emit_queue.task_done()

  async def emit(self, evt, data, priority=0, **kwargs):
    await self._emit_queue.put((
      priority, next(self._emit_counter), 
      (evt, data),
      kwargs,
    ))

  async def disconnect(self):
    if self._emit_enabled.is_set():
      logger.debug('Ensuring emit queue has been fully processed...')
      await self._emit_queue.join()
    await super().disconnect()



import uuid
import traceback
import socketio
import logging
logger = logging.getLogger(__name__)

from appyter.ext.socketio.forwarding import ForwardingMixin
from appyter.ext.socketio.priority_queued_emit import PriorityQueuedEmitMixin
from appyter.ext.socketio.chunked_emit import ChunkedEmitMixin

class AsyncClient(
  ChunkedEmitMixin,
  ForwardingMixin,
  PriorityQueuedEmitMixin,
  socketio.AsyncClient,
):
  ''' Client with packet forwarding, emit buffering & chunking
   `async with` is necessary for initialization
  '''
  async def __aenter__(self):
    await super().__aenter__()
    #
    @self.event
    async def connect():
      self._emit_enabled.set()
    @self.event
    async def disconnect():
      self._emit_enabled.clear()
    #
    return self

  async def __aexit__(self, *args):
    await super().__aexit__(*args)
    try: await self.eio.http.close()
    except: logger.warn(traceback.format_exc())

  async def connect(self, *args, headers={}, **kwargs):
    return await super().connect(
      *args,
      headers=dict({ 'Authorization': str(uuid.uuid4()) }, **headers),
      **kwargs,
    )



''' Manage rooms
NOTE: `rooms` state would need to be distributed if the
server was distributed.
'''
import uuid
import contextlib
import logging
logger = logging.getLogger(__name__)

from appyter.render.flask_app.socketio import socketio

rooms = {}

async def ensure_room(room):
  global rooms
  if room not in rooms:
    logger.debug(f"creating room {room}")
    rooms[room] = {
      'users': {},
      'clients': set(),
      'messages': [],
    }
  return rooms[room]

async def remove_client_from_room(room_id, client_sid):
  global rooms
  if room_id in rooms:
    rooms[room_id]['clients'].remove(client_sid)
    if not rooms[room_id]['clients']:
      # no more clients, close room
      logger.debug(f"closing room {room_id}")
      del rooms[room_id]

async def enter_room(client_sid, room_id):
  room = await ensure_room(room_id)
  if client_sid not in room['clients']:
    logger.debug(f"client {client_sid} joined room {room_id}")
    # add client to room and relay old messages to the client
    room['clients'].add(client_sid)
    await socketio.enter_room(client_sid, room_id)
    async with socketio.session(client_sid) as client_sess:
      client_uid = client_sess['uid']
    for msg in room['messages'][room['users'].get(client_uid, 0):]:
      await socketio.emit(msg['event'], msg['data'], to=client_sid)
    room['users'][client_uid] = len(room['messages'])

async def ensure_joined(client_sid):
  async with socketio.session(client_sid) as client_sess:
    client_uid = client_sess['uid']
  #
  for room_id in rooms:
    if client_uid in rooms[room_id]['users']:
      await enter_room(client_sid, room_id)

async def find_room(room_id):
  global rooms
  return room_id in rooms

@contextlib.asynccontextmanager
async def room_lock(room_id):
  lock_id = str(uuid.uuid4())
  room = await ensure_room(room_id)
  room['clients'].add(lock_id)
  try:
    yield
  finally:
    await remove_client_from_room(room_id, lock_id)


@socketio.on('join')
async def on_join(client_sid, room_data):
  if type(room_data) == str: room_data = {'_id': room_data}
  await enter_room(client_sid, room_data['_id'])

@socketio.on('forward')
async def forward(client_sid, data):
  global rooms
  if data['to'] in rooms:
    rooms[data['to']]['messages'].append(data)

@socketio.on('disconnect')
async def _(client_sid):
  for room in socketio.rooms(client_sid):
    await remove_client_from_room(room, client_sid)



import re
import aiohttp
from appyter.ext.fsspec.core import url_to_fs_ex
from appyter.ext.asyncio.event_loop import get_event_loop
from fsspec.asyn import AsyncFileSystem
from appyter.ext.fsspec.spec.sync_async import SyncAsyncFileSystem
from appyter.ext.urllib import URI, join_url

drs_matcher = re.compile(r'^drs://([^/]+)/(.+)$', re.IGNORECASE)

class DRSFileSystem(SyncAsyncFileSystem, AsyncFileSystem):
  protocol = 'drs'

  def __init__(self, *args, ssl_verify=True, **kwargs):
    super().__init__(*args, ssl_verify=ssl_verify, **kwargs)
    self.ssl_verify = ssl_verify

  @classmethod
  def _strip_protocol(cls, path):
    """For DRS, we always want to keep the full URL"""
    return path

  @classmethod
  def _parent(cls, path):
    # override, since _strip_protocol is different for URLs
    par = super()._parent(path)
    if len(par) > 6:  # "drs://..."
        return par
    return ""

  async def __aenter__(self):
    self._session_mgr = aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=self.ssl_verify))
    self._session = await self._session_mgr.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self._session_mgr.__aexit__(exc_type, exc_value, traceback)

  async def _info(self, path, **kwargs):
    ''' Use DRS To find info about the object
    _id contains the DRS ID
    _url contains an fsspec compatible url
    '''
    # lookup DRS object
    async with self._session.get(drs_matcher.sub(r'https://\1/ga4gh/drs/v1/objects/\2', path)) as req:
      if req.status != 200:
        raise FileNotFoundError
      info = await req.json()
    # obtain first compatible access method
    acceptible_types = {'s3', 'gs', 'ftp', 'https'}
    access_method = None
    for access_method in info.get('access_methods', []):
      if access_method['type'] in acceptible_types:
        break
    if access_method is None: raise RuntimeError('Invalid DRS Implementation')
    elif access_method['type'] not in acceptible_types: raise NotImplementedError
    # obtain access url
    if access_method.get('access_url'):
      access_url = dict(url=access_method['access_url'])
    elif access_method.get('access_id'):
      async with self._session.get(join_url(
        drs_matcher.sub(r'https://\1/ga4gh/drs/v1/objects/\2/access/', path),
        access_method['access_id'],
      )) as req:
        if req.status != 200:
          raise RuntimeError('Invalid DRS Implementation')
        access_url = await req.json()
    else:
      raise RuntimeError('Invalid DRS Implementation')
    # convert access_url into fsspec url
    url = URI(access_url['url']).with_fragment(None)
    if access_url.get('headers'):
      url = url.with_fragment_query(access_url.get('headers', {}))
    if access_method.get('region'):
      url = url.update_fragment_query({
        f"{access_method['type']}.region": access_method['region']
      })
    #
    return {
      '_id': info['id'],
      '_url': str(url),
      'type': 'file',
      'name': path,
      'size': info['size'],
    }

  async def _ls(self, path, **kwargs):
    raise NotImplementedError

  def _open(self, path, mode="rb", **kwargs):
    ''' Open the path with the url prepared during `_info`
    '''
    if mode != "rb": raise NotImplementedError
    path_info = self.info(path)
    kwargs = dict(loop=get_event_loop())
    if path_info['_url'].startswith('https://'):
      kwargs.update(client_kwargs=dict(connector=aiohttp.TCPConnector(ssl=self.ssl_verify, loop=kwargs['loop'])))
    fs, fo = url_to_fs_ex(path_info['_url'], **kwargs)
    return fs.open(fo, mode)



import re
from appyter.extras.catalog_integration.user_config import get_user_config

async def prepare_executor(data):
  ''' Safely translate public facing executor URI into internally operable execspec URIs
  '''
  # if executor hints are provided (e.g. ['executor:default'])
  #  we'll verify that the chosen executor is supported
  executor_hints = {
    hint_value
    for hint in data.get('_config').get('HINTS', set())
    for hint_prefix, sep, hint_value in (hint.partition(':'),)
    if hint_prefix == 'executor' and sep == ':'
  }
  if executor_hints and data.get('_executor', 'default') not in executor_hints:
    raise Exception('Executor not supported for this appyter')
  #
  if data.get('_executor'):
    if data['_executor'] == 'cavatica':
      user_config = await get_user_config(data.get('_auth'), data.get('_config'))
      if not user_config.get('cavatica_api_key'): raise PermissionError
      assert re.match(r'^[0-9a-f]+$', user_config['cavatica_api_key']) is not None, 'Malformed CAVATICA API Key'
      m = re.match(r'^cavatica://(?P<user>[a-z0-9_-]+)/(?P<project>[a-z0-9_-]+)$', data['_storage'])
      if m:
        return f"cavatica#?cwl=appyter.cwl&project={m.group('user')}/{m.group('project')}&auth_token={user_config['cavatica_api_key']}"
      else:
        raise Exception('Unsupported storage URI')



import os
import click
import logging
from appyter import __version__
from appyter.cli import cli
from appyter.ext.click import click_option_setenv, click_argument_setenv

logger = logging.getLogger(__name__)

@cli.command(
  help='An alias for jupyter notebook which pre-fetches notebook data from a remote appyter and then launches serve',
  context_settings=dict(
    ignore_unknown_options=True,
  ),
)
@click_option_setenv('--data-dir', envvar='APPYTER_DATA_DIR', default=None, help='The directory to store data of executions')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_option_setenv('--host', envvar='APPYTER_HOST', default='0.0.0.0', help='The host to bind to')
@click_option_setenv('--port', envvar='APPYTER_PORT', type=int, default=5000, help='The port this server should run on')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
@click_argument_setenv('uri', envvar='APPYTER_URI', nargs=1, type=str)
@click.pass_context
def fetch_and_serve(ctx, data_dir, cwd, host, port, args, uri):
  import urllib.parse, urllib.request
  # TODO: can this be less reliant on the `appyter-catalog` storage setup
  uri_parsed = urllib.parse.urlparse(uri)
  from appyter.ext.fsspec.storage import ensure_storage
  from appyter.ext.asyncio.helpers import ensure_sync_contextmanager
  from appyter.ext.fsspec.core import url_to_chroot_fs
  from appyter.ext.asyncio.event_loop import with_event_loop
  with with_event_loop():
    with ensure_sync_contextmanager(ensure_storage(f"{uri_parsed.scheme}://{uri_parsed.netloc}/storage/appyters/")):
      # if data_dir doesn't exist, create it
      if data_dir is None: data_dir = 'memory://'
      # mount the appyter into the data_dir
      fs = url_to_chroot_fs(data_dir, appyter=uri)
      with fs.mount(fuse=False) as mnt:
        logging.info(f"Starting `appyter serve`...")
        # serve the bundle in jupyter notebook
        from appyter.helpers.serve import serve
        ctx.invoke(serve, cwd=cwd, data_dir=str(mnt), host=host, port=port, args=args)



import re
from appyter.context import get_env

from appyter.ext.flask import request_get
from appyter.ext.dict import dict_filter_none

async def prepare_request(req):
  ''' Extract additional fields from request
  '''
  data = dict(_config=get_env())
  data.update(dict_filter_none({
    k: request_get(req, k)
    for k in [
      '_auth',
      '_executor',
      '_storage',
    ]
  }))
  #
  if not data.get('_auth') and getattr(req, 'cookies', None) is not None:
    authorization = req.cookies.get('authorization')
    if authorization:
      data['_auth'] = authorization

  if not data.get('_auth') and getattr(req, 'headers', None) is not None:
    authorization = req.headers.get('Authorization')
    if authorization:
      m = re.match(r'^Bearer (.+)$', authorization)
      if m:
        data['_auth'] = m.group(1)
  #
  if not data.get('_storage') and getattr(req, 'args', None) is not None:
    storage = req.args.get('storage')
    if storage:
      data['_storage'] = storage
  #
  if data.get('_executor') == 'cavatica':
    # CAVATICA executor must be authenticated
    if not data.get('_auth'): raise PermissionError
    # CAVATICA executor should use cavatica storage
    if not data.get('_storage'):
      # NOTE: this gets sanitized downstream
      cavatica_project = request_get(req, '_cavatica_project')
      if not cavatica_project:
        from appyter.extras.catalog_integration.user_config import get_user_config
        user_config = await get_user_config(data.get('_auth'), data.get('_config'))
        if not user_config.get('cavatica_api_key'): raise Exception('Missing CAVATICA API Key')
        if not user_config.get('cavatica_project'): raise Exception('Missing CAVATICA Project')
        cavatica_project = user_config.get('cavatica_project')
      data['_storage'] = f"cavatica://{cavatica_project}"
    else:
      assert data['_storage'].startsWith('cavatica://'), 'CAVATICA executor requires CAVATICA storage'
  #
  return data



import contextlib
import traceback
import logging

logger = logging.getLogger(__name__)

from appyter.context import get_env, get_jinja2_env
from appyter.ext.dict import dict_collision_free_update
from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.parse.nb import nb_to_ipynb_io
from appyter.render.flask_app.constants import get_fields, get_ipynb_hash, get_nbtemplate
from appyter.render.nbconstruct import render_nb_from_nbtemplate
from appyter.ext.hashlib import sha1sum_dict
from appyter.ext.asyncio.helpers import ensure_async, ensure_async_contextmanager, ensure_sync
from appyter.ext.urllib import URI

async def _prepare_request(req):
  ''' Extract internal session variables from request
  '''
  data = dict(_config=get_env())
  if 'catalog-integration' in data['_config']['EXTRAS']:
    from appyter.extras.catalog_integration.request import prepare_request as prepare_request_catalog
    data.update(await prepare_request_catalog(req))
  return data
prepare_request = ensure_sync(_prepare_request)

def prepare_data(req):
  ''' Extract fields from submission
  '''
  data = {}
  for field in get_fields():
    dict_collision_free_update(data, **field.prepare(req))
  return data
_prepare_data = ensure_async(prepare_data)

async def _prepare_storage(data):
  ''' Figure out the storage fsspec URI to use for this request
  '''
  storage_uri = None
  #
  if 'catalog-integration' in data['_config']['EXTRAS']:
    from appyter.extras.catalog_integration.storage import prepare_storage as prepare_storage_catalog
    storage_uri = await prepare_storage_catalog(data)
  #
  if storage_uri is None:
    storage_uri = data['_config']['DATA_DIR']#'storage://'
  #
  return storage_uri
prepare_storage = ensure_sync(_prepare_storage)

async def _prepare_results(data):
  ''' Compute instance id & ensure results exist in storage
  '''
  storage = await _prepare_storage(data)
  if 'catalog-integration' in data['_config']['EXTRAS']:
    from appyter.extras.catalog_integration.userfs import upload_user_to_storage
    data = await upload_user_to_storage(storage, data)
  instance_id = sha1sum_dict(dict(
    ipynb=get_ipynb_hash(),
    data={
      k: v
      for k, v in data.items()
      if not k.startswith('_') and k not in {'_executor', '_storage'}
    },
  ))
  cwd = str(URI(storage).join('output', instance_id))
  data_fs = url_to_chroot_fs(cwd)
  data_fs_ctx = ensure_async_contextmanager(data_fs)
  async with data_fs_ctx:
    await ensure_async(data_fs.makedirs)('', exist_ok=True)
    if not await ensure_async(data_fs.exists)(data['_config']['IPYNB']):
      # construct notebook
      env = get_jinja2_env(config=data['_config'], context=data, session=instance_id)
      nbtemplate = get_nbtemplate()
      # in case of constraint failures, we'll fail here
      nb = render_nb_from_nbtemplate(env, nbtemplate, data=data)
      # write notebook
      async with ensure_async_contextmanager(
        await ensure_async(data_fs.open)(data['_config']['IPYNB'], 'w')
      ) as fw:
        await ensure_async(nb_to_ipynb_io)(nb, fw)
      #
      if 'catalog-integration' in data['_config']['EXTRAS']:
        # if you create a notebook, it should get registered
        try:
          from appyter.extras.catalog_integration.notebooks import InstanceInfo, add_instance
          await add_instance(
            InstanceInfo(
              instance=instance_id,
              metadata=dict(
                appyter=nb.metadata.get('appyter'),
                storage=data.get('_storage'),
                executor=data.get('_executor'),
              )
            ),
            auth=data.get('_auth'),
            config=data.get('_config'),
          )
        except:
          logger.warning(traceback.format_exc())
      return instance_id
  #
  if 'catalog-integration' in data['_config']['EXTRAS']:
    # if you recreate a notebook, it should get registered,
    #  we'll omit metadata which would already be captured
    #  when the instance was created
    try:
      from appyter.extras.catalog_integration.notebooks import InstanceInfo, add_instance
      await add_instance(
        InstanceInfo(instance=instance_id),
        auth=data.get('_auth'),
        config=data.get('_config'),
      )
    except:
      logger.warning(traceback.format_exc())
  #
  return instance_id
prepare_results = ensure_sync(_prepare_results)

@contextlib.asynccontextmanager
async def _prepare_executor(data, fallback=None):
  ''' Figure out the executor to use for this request
  '''
  executor_uri = None
  #
  if 'catalog-integration' in data['_config']['EXTRAS']:
    from appyter.extras.catalog_integration.executor import prepare_executor as prepare_executor_catalog
    executor_uri = await prepare_executor_catalog(data)
  #
  if executor_uri is not None:
    from appyter.execspec.core import url_to_executor
    async with url_to_executor(executor_uri, config=data['_config']) as executor:
      yield executor
  else:
    yield fallback
prepare_executor = ensure_sync(_prepare_executor)



import os
import json
import click
import fsspec

from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.urllib import join_slash
from appyter.parse.nb import nb_from_ipynb_io
from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
from appyter.render.nbinspect.cli import nbinspect
from appyter.ext.dict import dict_filter_none

def render_jsonschema_from_nbtemplate(env, nb):
  ''' Render a jsonschema representing the relevant Fields throughout the notebook.
  TODO: branch with tabs
  '''
  fields = list(parse_fields_from_nbtemplate(env, nb, deep=True))
  schema = {
    'type': 'object',
    'properties': dict_filter_none({
      field.args['name']: field.to_jsonschema()
      for field in fields
    }),
  }
  required = [
    field.args['name']
    for field in fields
    if field.args.get('required')
  ]
  if required: schema['required'] = required
  return schema

@nbinspect.command(help='Create JSON Schema for appyter input')
@click.option('-o', '--output', default='-', type=click.File('w'), help='The output location of the inspection json')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def jsonschema(cwd, ipynb, output, **kwargs):
  cwd = os.path.realpath(cwd)
  env = get_jinja2_env(config=get_env(cwd=cwd, ipynb=ipynb, mode='inspect', **kwargs))
  with fsspec.open(join_slash(cwd, ipynb), 'r') as fr:
    nbtemplate = nb_from_ipynb_io(fr)
  json.dump(render_jsonschema_from_nbtemplate(env, nbtemplate), output)



def dict_filter_none(d):
  return { k: v for k, v in d.items() if v is not None }

def dict_collision_free_update(d, **kwargs):
  for k, v in kwargs.items():
    assert k not in d, 'Collision in update'
    d[k] = v

def dict_flatten(d):
  Q = [(tuple(), d)]
  F = []
  while Q:
    K, V = Q.pop()
    if type(V) == dict:
      for k, v in V.items():
        Q.append(((*K, k), v))
    else:
      F.append((K, V))
  return F

def dict_unflatten(f):
  D = {}
  for K, v in f:
    o = D
    for k in K[:-1]:
      if k not in o:
        o[k] = {}
      o = o[k]
    o[K[-1]] = v
  return D

def dict_merge(d, **kwargs):
  return dict_unflatten(dict_flatten(d) + dict_flatten(kwargs))

def expand_dotmap(dotmap):
  ''' Convert dot-notation into deep dict
  '''
  params = {}
  for K, v in dotmap.items():
    _params_n_2 = None
    _params_n_1 = params
    for k in K.split('.'):
      if k not in _params_n_1: _params_n_1[k] = {}
      _params_n_2 = _params_n_1
      _params_n_1 = _params_n_2[k]
    _params_n_2[k] = v
  return params

def dict_typed_flatten(d):
  ''' dict_flatten but preserves lists & ordering
  '''
  Q = [(tuple(), d)]
  F = []
  while Q:
    K, V = Q.pop()
    if type(V) == dict:
      for k, v in V.items():
        Q.append(((*K, dict, k), v))
    elif type(V) == list:
      for k, v in enumerate(V):
        Q.append(((*K, list, k), v))
    else:
      F.append((K, V))
  return F

def dict_typed_unflatten(f):
  ''' opposite of dict_typed_unflatten
  '''
  D = {}
  for K, v in f:
    o = D
    for k, typ in zip(K[1:-1:2], K[2:-1:2]):
      if type(o) == dict and k not in o:
        o[k] = typ()
      elif type(o) == list and len(o) < k or o[k] is None:
        while len(o) <= k: o.append(None)
        o[k] = typ()
      #
      if type(o) == list:
        while len(o) <= k: o.append(None)
      o = o[k]
    if type(o) == list:
      while len(o) <= K[-1]: o.append(None)
    o[K[-1]] = v
  return D



import contextlib
import logging
import multiprocessing as mp

logger = logging.getLogger(__name__)

def _fuse_run(fs_json, fs_path, mount_dir):
  import fsspec.fuse
  import appyter.ext.fsspec
  from appyter.ext.asyncio.event_loop import with_event_loop
  from appyter.ext.fsspec.spec.composable import ComposableAbstractFileSystem
  with with_event_loop():
    fs = ComposableAbstractFileSystem.from_json(fs_json)
    logger.debug(f'preparing {fs}..')
    with fs as fs:
      logger.debug('launching fuse..')
      fsspec.fuse.run(fs, fs_path, mount_dir)
      logger.debug('teardown..')

@contextlib.asynccontextmanager
async def fs_mount(fs, fs_path='', mount_dir=None):
  import os
  import signal
  import traceback
  from appyter.ext.asyncio.try_n_times import async_try_n_times
  from appyter.ext.asyncio.helpers import ensure_async
  from appyter.ext.tempfile import tempdir
  def assert_true(val): assert val
  with tempdir(mount_dir) as mount_dir:
    logger.debug(f'mounting {fs} onto {mount_dir}')
    proc = mp.Process(
      target=_fuse_run,
      args=(fs.to_json(), str(fs_path), str(mount_dir)),
    )
    proc.start()
    try:
      await async_try_n_times(3, ensure_async(lambda path: assert_true(path.is_mount())), mount_dir)
      logger.debug(f"fs mount ready on {mount_dir}")
      yield mount_dir
    except Exception as e:
      logger.error(traceback.format_exc())
      raise
    finally:
      if proc.pid:
        logger.debug(f"unmounting fs from {mount_dir}")
        os.kill(proc.pid, signal.SIGINT) # SIGINT cleanly stops fsspec.fuse.run
        logger.debug(f"waiting for process to end")
        try:
          await ensure_async(proc.join)(30)
        except TimeoutError:
          os.kill(proc.pid, signal.SIGTERM) # SIGTERM unclean stop
        await async_try_n_times(3, ensure_async(lambda path: assert_true(not path.is_mount())), mount_dir)
    logger.debug(f"done")



import os
import io
import re
import random
import asyncio
import json
import aiohttp
import logging
import traceback
import functools

from appyter.parse.nb import nb_from_ipynb_io
logger = logging.getLogger(__name__)

from appyter.execspec.spec import AbstractExecutor
from appyter.ext.fsspec.util import fsspec_read_and_run
from appyter.ext.asyncio.try_n_times import async_try_n_times
from appyter.ext.dict import dict_filter_none, dict_merge
from appyter.ext.urllib import URI, join_slash

class WESExecutor(AbstractExecutor):
  ''' Run executions via a workflow execution service endpoint
  '''
  protocol = 'wes'

  def __init__(self, url=None, config={}, **kwargs):
    super().__init__(url=re.sub(r'^wes://', 'https://', url), **kwargs)
    self.config = config

  async def __aenter__(self):
    await super().__aenter__()
    from appyter.parse.nb import nb_from_ipynb_io
    self.cwl = await fsspec_read_and_run(self.executor_options['cwl'], lambda fr: json.load(fr))
    self.nbtemplate = await fsspec_read_and_run(join_slash(self.config['CWD'], self.config['IPYNB']), nb_from_ipynb_io)
    self.session = aiohttp.ClientSession(
      headers=dict(
        {'Accept': 'application/json'},
        **self.executor_options.get('headers', {}),
      )
    )
    self.client = await self.session.__aenter__()
    return self
  
  async def __aexit__(self, type, value, traceback):
    await self.session.__aexit__(type, value, traceback)
    await super().__aexit__(type, value, traceback)

  async def _prepare(self, job):
    # NOTE: This is redundant, since the notebook was already created
    #       but is necessary if we want to use the standard `cwl-runner`
    from appyter.context import get_jinja2_env
    from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
    nb = await fsspec_read_and_run(str(URI(job['storage']).join(job['cwd'], job['ipynb'])), nb_from_ipynb_io)
    env = get_jinja2_env(config=self.config, context=nb.metadata['appyter']['nbconstruct']['data'], session=job['session'])
    inputs = dict_filter_none({
      field.args['name']: field.to_cwl_value()
      for field in parse_fields_from_nbtemplate(env, self.nbtemplate, deep=True)
    })
    inputs['s'] = job['url']
    inputs['w'] = f"appyter/output/{job['session']}"
    return dict_merge(
      self.executor_options.get('params', {}),
      workflow_params=dict(inputs=inputs),
      workflow_url='#/workflow_attachment/0',
      workflow_attachment=[
        [
          os.path.basename(self.executor_options['cwl']),
          functools.partial(io.BytesIO, json.dumps(self.cwl).encode()),
        ],
      ],
      workflow_type='CWL',
      workflow_type_version=self.cwl['cwlVersion'],
    )

  async def _submit(self, wes_job):
    data = aiohttp.FormData()
    for k, v in wes_job.items():
      if k == 'workflow_attachment':
        for filename, bytesio in v:
          data.add_field('workflow_attachment', bytesio(), filename=filename)
      elif type(v) in {dict, list}:
        data.add_field(k, json.dumps(v), content_type='application/json')
      else:
        data.add_field(k, v)
    async with self.client.post(
      join_slash(self.url, 'ga4gh/wes/v1', 'runs'),
      data=data,
    ) as req:
      res = await req.json()
      logger.info(f"{res=}")
      return res['run_id']

  async def _watch_state(self, run_id):
    current_state = None
    while True:
      await asyncio.sleep(15 * (0.5 + random.random()))
      logger.debug(f"Checking status of job {run_id=}")
      async with self.client.get(join_slash(self.url, 'ga4gh/wes/v1', 'runs', run_id, 'status')) as req:
        res = await req.json()
        state = res['state']
      logger.debug(f"{state=}")
      if current_state != state:
        current_state = state
        yield current_state

  async def _finalize(self, job, wes_job):
    pass

  async def _run(self, **job):
    yield dict(type='status', data=f"Preparing WES parameters...")
    wes_job = await async_try_n_times(3, self._prepare, job)
    yield dict(type='status', data=f"Dispatching WES execution...")
    logger.info(f"{wes_job=}")
    run_id = await async_try_n_times(3, self._submit, wes_job)
    yield dict(type='status', data=f"Dispatched successfully, your execution will begin soon..")
    async for state in self._watch_state(run_id):
      if state == 'QUEUED':
        yield dict(type='status', data=f"Queued successfully..")
      elif state == 'INITIALIZING':
        yield dict(type='status', data=f"Computational resources are being initialized..")
      elif state == 'RUNNING':
        yield dict(type='status', data=f"Task is now running..")
      elif state == 'CANCELING':
        yield dict(type='error', data=f"Task is being cancelled!")
        break
      elif state == 'CANCELED':
        yield dict(type='error', data=f"Task was cancelled!")
        break
      elif state == 'EXECUTOR_ERROR':
        yield dict(type='error', data=f"Task had an error and did not complete successfully!")
        break
      elif state == 'COMPLETE':
        break
      else:
        yield dict(type='error', data=f"An unknown state reported by WES")
        logger.warning(f"Unhandled state received: {state}")
        break
    #
    try:
      await self._finalize(job, wes_job)
    except:
      logger.error(traceback.print_exc())
      yield dict(type='error', data=f"Task failed to finalize")
    else:
      yield dict(type='status', data=f"Task completed")



import appyter.ext.multiprocessing

import tempfile
import contextlib
from pathlib import Path

from appyter.ext.fsspec.mapperfs import MapperFileSystem

import pytest
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_mapperfs():
  with _test_ctx() as tmpdir:
    with MapperFileSystem(pathmap={'a/b/c': str(tmpdir/'a'/'b'/'c'), 'E': str(tmpdir/'e')}) as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'E']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('E'), b'E')



import contextlib
import traceback
import logging
logger = logging.getLogger(__name__)

def json_emitter_factory(output):
  async def json_emitter(obj):
    import json
    print(json.dumps(obj), file=output)
  return json_emitter

@contextlib.asynccontextmanager
async def url_to_emitter(url):
  if url.startswith('file://'):
    import fsspec
    with fsspec.open(url, 'w') as fw:
      yield json_emitter_factory(fw)
  elif url.startswith('http://') or url.startswith('https://'):
    import fsspec
    import urllib.parse
    from appyter.ext.socketio import AsyncClient
    with fsspec.open('file:///dev/stderr', 'w') as fw:
      fallback_emitter = json_emitter_factory(fw)
      async with AsyncClient() as sio:
        url_parsed = urllib.parse.urlparse(url)
        url_path, room = url_parsed.path.rsplit('/', maxsplit=1)
        try:
          await sio.connect(f"{url_parsed.scheme}://{url_parsed.netloc}", socketio_path=url_path)
        except:
          await fallback_emitter(dict(
            type='error',
            data=f"Failed to connect to {url}... falling back to stderr",
            traceback=traceback.format_exc(),
          ))
        else:
          try:
            await sio.emit('join', room)
          except:
            await fallback_emitter(dict(
              type='error',
              data=f"Failed to join {room}...",
              traceback=traceback.format_exc(),
            ))
        async def emit(msg):
          try:
            await sio.emit(msg['type'], msg['data'], to=room)
          except:
            await fallback_emitter(dict(msg,
              traceback=traceback.format_exc(),
            ))
        try:
          yield emit
        finally:
          try:
            await sio.disconnect()
          except:
            await fallback_emitter(dict(
              type='error',
              data=f"Failed to disconnect.",
              traceback=traceback.format_exc(),
            ))
  else:
    logger.error("Emitter not recognized, falling back to stderr")
    import fsspec
    with fsspec.open('file:///dev/stderr', 'w') as fw:
      yield json_emitter_factory(fw)



import logging
logger = logging.getLogger(__name__)
from dataclasses import dataclass, asdict
from appyter.ext.urllib import join_url, parent_url

@dataclass(frozen=True,)
class InstanceInfo:
  instance: str
  metadata: dict = None

async def add_instance(data: InstanceInfo, auth=None, config=None):
  import aiohttp
  async with aiohttp.ClientSession(
    headers={
      'Authorization': f"Bearer {auth}",
    } if auth else {},
  ) as session:
    async with session.post(
      join_url(
        parent_url(config['PUBLIC_URL']),
        'postgrest/rpc/add_instance',
      ),
      json=asdict(data),
    ) as res:
      if res.status != 200:
        raise Exception(await res.text())
      return await res.json()



import click
import yaml
from appyter import __version__
import logging.config as loggingConf
from click_default_group import DefaultGroup
from appyter.ext.click import click_option_setenv
from appyter.ext.dict import dict_merge

@click.group(cls=DefaultGroup, default='flask-app')
@click.version_option(
  prog_name='appyter',
  version=__version__,
)
@click_option_setenv('--debug', envvar='APPYTER_DEBUG', default=True, help='Debug or production')
@click_option_setenv('--logging', envvar='APPYTER_LOGGING', default='{}', help='Appyter logging config and/or file')
def cli(logging=None, debug=False):
  loggingConf.dictConfig(dict_merge({
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
      'standard': {
        'format': '%(levelname)s:%(name)s: %(message)s',
      },
    },
    'handlers': {
      'console': {
        'level': 'DEBUG' if debug else 'INFO',
        'formatter': 'standard',
        'class': 'logging.StreamHandler',
        'stream': 'ext://sys.stderr',
      },
    },
    'loggers': {
      '': {
        'handlers': ['console'],
        'level': 'INFO',
        'propagate': False,
      },
      __package__: {
        'handlers': ['console'],
        'level': 'DEBUG',
        'propagate': False,
      },
    },
  }, **yaml.safe_load(logging)))



''' Endpoints for exporting results
'''
import io
import shutil
import zipfile
from flask import request, current_app, send_file, abort

from appyter.ext.fsspec.core import url_to_chroot_fs
from appyter.render.flask_app.constants import get_base_files, get_html_exporer
from appyter.render.flask_app.core import core
from appyter.render.flask_app.prepare import prepare_storage, prepare_request
from appyter.ext.urllib import URI
from appyter.ext.flask import route_join_with_or_without_slash
from appyter.parse.nb import nb_from_ipynb_io

@route_join_with_or_without_slash(core, 'export', '<path:path>', methods=['GET'])
def export(path):
  if path.endswith('/'):
    data = dict(_id=path, _config=current_app.config)
    data.update(prepare_request(data))
    format = request.args.get('format', 'html')
    storage = URI(prepare_storage(data))
    with url_to_chroot_fs(str(storage.join('output', path))) as data_fs:
      nbpath = data['_config']['IPYNB']
      if data_fs.exists(nbpath):
        with data_fs.open(nbpath, 'rb') as fr:
          nb = nb_from_ipynb_io(fr)
        if format == 'html':
          exporter = get_html_exporer()
          body, _rcs = exporter.from_notebook_node(nb)
          return send_file(io.BytesIO(body.encode()), mimetype='text/html', as_attachment=True, download_name='output.html')
        elif format == 'zip':
          metadata = nb.get('metadata', {}).get('appyter', {})
          files = metadata.get('nbexecute', {}).get('files', metadata.get('nbconstruct', {}).get('files', {}))
          with url_to_chroot_fs('memory://') as tmp_fs:
            with tmp_fs.open('output.zip', 'wb') as fw:
              with zipfile.ZipFile(fw, 'a', zipfile.ZIP_DEFLATED, False) as zf:
                for f, b in get_base_files().items():
                  zf.writestr(f, b)
                for f in (nbpath, *files):
                  if data_fs.exists(f):
                    with data_fs.open(f, 'rb') as fr:
                      with zf.open(f, 'w') as zfw:
                        shutil.copyfileobj(fr, zfw)
            return send_file(tmp_fs.open('output.zip', 'rb'), mimetype='application/zip', as_attachment=True, download_name='output.zip')
  abort(404)



import fsspec

from appyter.ext.fsspec.writecache import WriteCacheFileSystem
fsspec.register_implementation('writecache', WriteCacheFileSystem)

from appyter.ext.fsspec.chroot import ChrootFileSystem
fsspec.register_implementation('chroot', ChrootFileSystem)

from appyter.ext.fsspec.sbfs import SBFSFileSystem
fsspec.register_implementation('sbfs', SBFSFileSystem)

from appyter.ext.fsspec.drs import DRSFileSystem
fsspec.register_implementation('drs', DRSFileSystem)

try:
  from appyter.ext.fsspec.s3fs import S3FileSystemEx
  fsspec.register_implementation('s3', S3FileSystemEx)
except ImportError:
  pass



import logging
logger = logging.getLogger(__name__)

import json
import aiohttp
from pathlib import Path
from fsspec.asyn import AsyncFileSystem
from appyter.ext.fsspec.spec.sync_async import SyncAsyncFileSystem
from appyter.ext.fsspec.spec.mountable import MountableAbstractFileSystem
from appyter.ext.asyncio.helpers import ensure_sync
from appyter.ext.urllib import join_slash
from fsspec.spec import AbstractBufferedFile

class SBFSFileSystem(MountableAbstractFileSystem, SyncAsyncFileSystem, AsyncFileSystem):
  CHUNK_SIZE = 8192

  def __init__(self, *args,
    api_endpoint='https://cavatica-api.sbgenomics.com',
    drs_endpoint='drs://cavatica-ga4gh-api.sbgenomics.com',
    auth_token=None,
    **storage_options,
  ):
    assert auth_token is not None, 'SBFS auth_token is required'
    super().__init__(*args,
      api_endpoint=api_endpoint,
      drs_endpoint=drs_endpoint,
      auth_token=auth_token,
      **storage_options,
    )
    self.auth_token = auth_token
    self.drs_endpoint = drs_endpoint
    self.api_endpoint = api_endpoint

  async def __aenter__(self):
    self._session_mgr = aiohttp.ClientSession(
      headers={
        'X-SBG-Auth-Token': self.auth_token,
      },
      raise_for_status=True,
    )
    self._session = await self._session_mgr.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self._session_mgr.__aexit__(exc_type, exc_value, traceback)

  async def _mkdir(self, path, create_parents=True, exist_ok=True, **kwargs):
    try:
      path_info = await self._info(path)
      if path_info['type'] == 'directory' and exist_ok:
        return path_info
      else:
        raise FileExistsError
    except FileNotFoundError:
      pass
    path_split = path.split('/')
    if len(path_split) <= 2: raise PermissionError
    if len(path_split) == 3:
      user, project, directory = path_split
      async with self._session.post(f"{self.api_endpoint}/v2/files", json=dict(
        project=f"{user}/{project}",
        name=directory,
        type='folder',
      ), raise_for_status=False) as req:
        if req.status == 404:
          raise NotADirectoryError
        else:
          req.raise_for_status()
        res = await req.json()
      return {
        '_id': res['id'],
        'name': f"{user}/{project}/{res['name']}",
        'type': 'directory',
      }
    else:
      parent_directory_path = '/'.join(path_split[:-1])
      directory = path_split[-1]
      try:
        parent_info = await self._info(parent_directory_path)
      except FileNotFoundError:
        if not create_parents:
          raise
        parent_info = await self._mkdir(parent_directory_path, create_parents=True, exist_ok=True)
      
      async with self._session.post(f"{self.api_endpoint}/v2/files", json=dict(
        parent=parent_info['_id'],
        name=directory,
        type='folder',
      )) as req:
        res = await req.json()
      return {
        '_id': res['id'],
        'name': f"{parent_directory_path}/{res['name']}",
        'type': 'directory',
      }

  async def _makedirs(self, path, exist_ok=False, **kwargs):
    return await self._mkdir(path, create_parents=True, exist_ok=exist_ok, **kwargs)

  async def _sbg_rm_file(self, file_id):
    async with self._session.delete(f"{self.api_endpoint}/v2/files/{file_id}", headers={
      'Content-Type': 'application/json',
    }) as res:
      return await res.text()

  async def _rmdir(self, path):
    info = await self._info(path)
    if info['type'] != 'directory': raise NotADirectoryError
    await self._sbg_rm_file(info['_id'])

  async def _rm_file(self, path):
    info = await self._info(path)
    if info['type'] == 'directory': raise IsADirectoryError
    await self._sbg_rm_file(info['_id'])

  async def _rm(self, path, recursive=False, **kargs):
    info = await self._info(path)
    if info['type'] == 'file':
      await self._sbg_rm_file(info['_id'])
    elif info['type'] == 'directory':
      if recursive:
        for child in await self._ls(info['name'], detail=False):
          await self._rm(child)
      await self._sbg_rm_file(info['_id'])
    else:
      raise NotImplementedError

  async def _mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    path1_info = await self._info(path1)
    path2_split = path2.split('/')
    if len(path2_split) <= 2: raise PermissionError
    project2 = '/'.join(path2_split[:2])
    if path1.startswith(project2): # moves only allowed within the same project
      try:
        path2_info = await self._info(path2)
        # overwrite/merge
        if path2_info['type'] == 'file':
          await self._rm_file(path2)
      except FileNotFoundError:
        # new file/directory
        if path1_info['type'] == 'directory':
          await self._mkdir(path2)
      #
      if path1_info['type'] == 'directory':
        if not recursive: raise IsADirectoryError
        for subpath1 in (await self._ls(path1, detail=False)):
          subpath2 = join_slash(path2, subpath1[len(path1):])
          await self._mv(subpath1, subpath2, recursive=recursive, maxdepth=(maxdepth-1) if maxdepth is not None else maxdepth, **kwargs)
        await self._rmdir(path1)
      elif path1_info['type'] == 'file':
        if await self._exists(path2):
          await self._rm_file(path2)
        if len(path2_split) == 3:
          async with self._session.post(f"{self.api_endpoint}/v2/files/{path1_info['_id']}/actions/move", json=dict(
            project=project2,
            name=path2_split[-1],
          )) as res:
            return await res.json()
        else:
          path2_parent_info = await self._info('/'.join(path2_split[:-1]))
          if path2_parent_info['type'] != 'directory': raise NotADirectoryError
          async with self._session.post(f"{self.api_endpoint}/v2/files/{path1_info['_id']}/actions/move", json=dict(
            parent=path2_parent_info['_id'],
            name=path2_split[-1],
          )) as res:
            return await res.json()
      else:
        raise NotImplementedError
    else:
      await self._cp_file(path1, path2, recursive=recursive, maxdepth=maxdepth)
      await self._rm(path1, recursive=recursive)

  async def _cp_file(self, path1, path2, **kwargs):
    path1_info = await self._info(path1)
    path2_split = path2.split('/', maxsplit=2)
    project2 = '/'.join(path2_split[:2])
    if len(path2_split) == 3:
      async with self._session.post(f"{self.api_endpoint}/v2/files/{path1_info['_id']}/actions/copy", json=dict(
        project=project2,
        name=path2_split[-1],
      )) as res:
        return await res.json()
    else:
      path2_parent_info = await self._info('/'.join(path2_split[:-1]))
      if path2_parent_info['type'] != 'directory': raise NotADirectoryError
      async with self._session.post(f"{self.api_endpoint}/v2/files/{path1_info['_id']}/actions/copy", json=dict(
        parent=path2_parent_info['_id'],
        name=path2_split[-1],
      )) as res:
        return await res.json()

  async def _download_info(self, file_info):
    async with self._session.get(f"{self.api_endpoint}/v2/files/{file_info['_id']}/download_info") as req:
      return await req.json()
  
  async def _cat_file(self, path, start=None, end=None, **kwargs):
    file_info = await self._info(path)
    download_info = await self._download_info(file_info)
    if end is None: end = file_info['size']
    elif end > file_info['size']: end = file_info['size']
    elif end < 0: raise Exception('Invalid end')
    if start is None: start = 0
    elif start >= end: return b''
    elif start < 0: raise Exception('Invalid start')
    async with self._session.get(download_info['url'], headers={
      'Range': f"bytes={start}-{end}"
    }) as req:
      return await req.read()

  async def _put_file(self, lpath, rpath, **kwargs):
    lpath = Path(lpath)
    lpath_stat = lpath.stat()
    rpath_split = rpath.split('/')
    if len(rpath_split) == 3:
      rpath_info = dict(
        project='/'.join(rpath_split[:2]),
        name=rpath_split[-1],
        size=lpath_stat.st_size,
        part_size=min(1073741824, lpath_stat.st_size),
      )
    elif len(rpath_split) > 3:
      parent_info = await self._info('/'.join(rpath_split[:-1]))
      rpath_info = dict(
        parent=parent_info['_id'],
        name=rpath_split[-1],
        size=lpath_stat.st_size,
        part_size=min(1073741824, lpath_stat.st_size),
      )
    else:
      raise PermissionError
    # initiate multipart upload
    logger.debug(f"Initiating multipart upload for {rpath}: {rpath_info}")
    async with self._session.post(f"{self.api_endpoint}/v2/upload/multipart", params=dict(overwrite='true'), json=rpath_info) as req:
      upload_info = await req.json()
    try:
      with lpath.open('rb') as fr:
        # for each chunk of size upload_info['part_size']
        for i in range((lpath_stat.st_size // upload_info['part_size']) + 1 if lpath_stat.st_size % upload_info['part_size'] else 0):
          # get part_info
          logger.info(f"Initiating multipart part {i+1} upload for {rpath_info['name']}")
          async with self._session.get(f"{self.api_endpoint}/v2/upload/multipart/{upload_info['upload_id']}/part/{i+1}") as req:
            part_info = await req.json()
          # load buffer with upload_info['part_size']
          buf = fr.read(upload_info['part_size'])
          # upload part
          async with self._session.request(part_info['method'], part_info['url'], data=buf) as req:
            await req.read()
            upload_response = {'headers': {k: json.loads(req.headers.get(k)) for k in part_info['report']['headers']}}
          logger.debug(f"{upload_response=}")
          # report uploaded part
          async with self._session.post(f"{self.api_endpoint}/v2/upload/multipart/{upload_info['upload_id']}/part", json=dict(
            part_number=i+1,
            response=upload_response,
          )) as req:
            await req.read()
      # report multipart completion
      async with self._session.post(f"{self.api_endpoint}/v2/upload/multipart/{upload_info['upload_id']}/complete", headers={
        'Content-Type': 'application/json',
      }) as req:
        await req.read()
    except:
      # abort multipart upload
      async with self._session.delete(f"{self.api_endpoint}/v2/upload/multipart/{upload_info['upload_id']}", headers={
        'Content-Type': 'application/json',
      }) as req:
        await req.read()
      raise
    logger.debug(f"Multipart upload for {rpath} completed")

  async def _get_file(self, rpath, lpath, **kwargs):
    file_info = await self._info(rpath)
    download_info = await self._download_info(file_info)
    lpath = Path(lpath)
    async with self._session.get(download_info['url']) as req:
      with lpath.open('wb') as fw:
        async for data in req.content.iter_chunked(SBFSFileSystem.CHUNK_SIZE):
          fw.write(data)

  async def _get_drs(self, path, **kwargs):
    file_info = await self._info(path, **kwargs)
    return f"{self.drs_endpoint}/{file_info['_id']}"

  def get_drs(self, path, **kwargs):
    return ensure_sync(self._get_drs(path, **kwargs))

  async def _info(self, path, **kwargs):
    if len(path) > 1024: raise FileNotFoundError
    path_split = [] if path in {'', '.', '/', './'} else path.split('/')
    if len(path_split) == 0:
      return {'name': path, 'type': 'directory'}
    elif len(path_split) == 1:
      project_user, = path_split
      async with self._session.get(f"{self.api_endpoint}/v2/projects/{project_user}") as req:
        projects = await req.json()
      if len(projects['items']) == 0:
        raise FileNotFoundError(path)
      elif len(projects['items']) > 1:
        raise Exception('Ambiguity')
      return {'name': project_user, 'type': 'directory'}
    elif len(path_split) == 2:
      project_user, proj_id = path_split
      async with self._session.get(f"{self.api_endpoint}/v2/projects/{project_user}", params=dict(
        name=proj_id
      ), raise_for_status=False) as req:
        if req.status == 404:
          raise FileNotFoundError(path)
        else:
          req.raise_for_status()
        projects = await req.json()
      if len(projects['items']) == 0:
        raise FileNotFoundError(path)
      elif len(projects['items']) == 2:
        raise Exception('Ambiguity')
      else:
        return {'name': f"{project_user}/{proj_id}", 'type': 'directory'}
    elif len(path_split) == 3:
      acc, proj_id, name = path_split
      proj = acc + '/' + proj_id
      async with self._session.get(f"{self.api_endpoint}/v2/files", params=dict(
        project=proj, name=name,
      ), raise_for_status=False) as req:
        if req.status == 404:
          raise FileNotFoundError(path)
        else:
          req.raise_for_status()
        items = await req.json()
      if len(items['items']) == 0:
        raise FileNotFoundError(path)
      elif len(items['items']) > 1:
        raise Exception('Ambiguity error')
      item = items['items'][0]
      item_name = f"{proj}/{item['name']}"
      if item['type'] == 'file':
        async with self._session.get(f"{self.api_endpoint}/v2/files/{item['id']}", params=dict(
          fields='size',
        )) as req:
          item_details = await req.json()
        return {
          '_id': item['id'],
          'name': item_name,
          'type': 'file',
          'size': item_details['size'],
        }
      elif item['type'] == 'folder':
        return {
          '_id': item['id'],
          'name': item_name,
          'type': 'directory',
        }
      else:
        raise NotImplementedError
    elif len(path_split) > 3:
      item_directory = '/'.join(path_split[:-1])
      parent_info = await self._info(item_directory)
      name = path_split[-1]
      async with self._session.get(f"{self.api_endpoint}/v2/files", params=dict(
        parent=parent_info['_id'], name=name,
      ), raise_for_status=False) as req:
        if req.status == 404:
          raise FileNotFoundError(path)
        else:
          req.raise_for_status()
        items = await req.json()
      if len(items['items']) == 0:
        raise FileNotFoundError(path)
      elif len(items['items']) > 1:
        raise Exception('Ambiguity error')
      item = items['items'][0]
      item_name = f"{item_directory}/{item['name']}"
      if item['type'] == 'file':
        async with self._session.get(f"{self.api_endpoint}/v2/files/{item['id']}", params=dict(
          fields='size',
        )) as req:
          item_details = await req.json()
        return {
          '_id': item['id'],
          'name': item_name,
          'type': 'file',
          'size': item_details['size'],
        }
      elif item['type'] == 'folder':
        return {
          '_id': item['id'],
          'name': item_name,
          'type': 'directory',
        }
      else:
        raise NotImplementedError

  async def _ls(self, path, detail=True, **kwargs):
    path_split = [] if path in {'', '.', '/', './'} else path.split('/')
    results = {}
    if len(path_split) == 0:
      async with self._session.get(f"{self.api_endpoint}/v2/projects") as req:
        projects = await req.json()
      for proj in projects['items']:
        proj_id_split = proj['id'].split('/')
        acc, _ = proj_id_split
        results[acc] = {'name': acc, 'type': 'directory'} if detail else acc
    elif len(path_split) == 1:
      project_user, = path_split
      async with self._session.get(f"{self.api_endpoint}/v2/projects/{project_user}") as req:
        projects = await req.json()
      for proj in projects['items']:
        proj_id_split = proj['id'].split('/')
        if path_split[0] == proj_id_split[0]:
          results[proj['id']] = {'name': proj['id'], 'type': 'directory'} if detail else proj['id']
    elif len(path_split) == 2:
      acc, proj_id = path_split
      proj = f"{acc}/{proj_id}"
      async with self._session.get(f"{self.api_endpoint}/v2/files", params=dict(
        project=proj,
      ), raise_for_status=False) as req:
        if req.status == 404:
          raise FileNotFoundError(path)
        else:
          req.raise_for_status()
        items = await req.json()
      for item in items['items']:
        item_name = proj + '/' + item['name']
        if detail:
          if item['type'] == 'file':
            async with self._session.get(f"{self.api_endpoint}/v2/files/{item['id']}", params=dict(
              fields='size',
            )) as req:
              file_details = await req.json()
            results[item_name] = {
              '_id': item['id'],
              'name': item_name,
              'type': 'file',
              'size': file_details['size'],
            }
          elif item['type'] == 'folder':
            results[item_name] = {
              '_id': item['id'],
              'name': item_name,
              'type': 'directory',
            }
          else:
            raise NotImplementedError
        else:
          results[item_name] = item_name
    elif len(path_split) > 2:
      path_info = await self._info(path)
      if path_info['type'] == 'directory':
        async with self._session.get(f"{self.api_endpoint}/v2/files", params=dict(
          parent=path_info['_id'],
        )) as req:
          items = await req.json()
        for item in items['items']:
          item_name = path_info['name'] + '/' + item['name']
          if detail:
            if item['type'] == 'file':
              async with self._session.get(f"{self.api_endpoint}/v2/files/{item['id']}", params=dict(
                fields='size',
              )) as req:
                file_details = await req.json()
              results[item_name] = {
                '_id': item['id'],
                'name': item_name,
                'type': 'file',
                'size': file_details['size'],
              }
            elif item['type'] == 'folder':
              results[item_name] = {
                '_id': item['id'],
                'name': item_name,
                'type': 'directory',
              }
            else:
              raise NotImplementedError
          else:
            results[item_name] = item_name
      else:
        return [path_info] if detail else [path_info['name']]
    return list(results.values())

  def _open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
    ''' Implements certain write ops, to use `open` with mode='w' or mode='a' use writecache,
    i.e. writecache::chroot::sbfs://*
    '''
    if mode != "rb":
      raise NotImplementedError
    return SBFSBufferedFile(self, path, mode=mode, block_size=block_size, cache_options=cache_options, **kwargs)

class SBFSBufferedFile(AbstractBufferedFile):
  def __init__(self, fs, path, mode="rb", **kwargs):
      super().__init__(fs, path, mode=mode, **kwargs)

  def _fetch_range(self, start, end):
    return self.fs.cat_file(self.path, start=start, end=end)



import os
import sys
import json
import click
import fsspec

from appyter import __version__
from appyter.context import get_env, get_jinja2_env
from appyter.ext.click import click_option_setenv, click_argument_setenv
from appyter.ext.urllib import join_slash
from appyter.parse.nb import nb_from_ipynb_io
from appyter.parse.nbtemplate import parse_fields_from_nbtemplate
from appyter.render.nbinspect.cli import nbinspect

from appyter.ext.dict import dict_filter_none
from appyter.parse.nbtemplate import parse_fields_from_nbtemplate

def render_cwl_from_nbtemplate(env, nb, ipynb=None, cwd=None, info=None):
  ''' Render a cwl representing the relevant Fields throughout the notebook.
  '''
  fields = list(parse_fields_from_nbtemplate(env, nb, deep=True))
  schema = {
    'cwlVersion': 'v1.2',
    'class': 'CommandLineTool',
    'requirements': [
        {'class': 'InlineJavascriptRequirement'},
        {'class': 'LoadListingRequirement'},
    ],
  }
  # potentially extract info from appyter embedded metadata if not provided
  if info is None and 'appyter' in nb.metadata and 'info' in nb.metadata['appyter']:
    info = nb.metadata['appyter']['info']
  if info is not None:
    schema.update({
      'id': f"appyter-{info['name'].lower()}",
      'baseCommand': [
        'python3', '-u',
        '-m', 'appyter',
        'commandify',
        f"/app/{ipynb}",
        "run"
      ],
      "hints": [
        {
          "class": "DockerRequirement",
          "dockerPull": f"maayanlab/appyter-{info['name'].lower()}:{info['version']}-{__version__}"
        }
      ],
      "label": info['title'],
      "doc": info['description'],
      "s:version": f"{info['version']}-{__version__}",
      "s:author": ', '.join(author['name'] for author in info['authors']),
      "s:license": info['license'],
      "s:codeRepository": info['url'],
      "s:keywords": info['tags'],
    })
  else:
    schema.update({
      'baseCommand': [
        sys.executable, '-u',
        '-m', 'appyter',
        'commandify',
        os.path.join(cwd, ipynb),
        "run"
      ],
    })
  #
  schema.update({
    'inputs': [
      *dict_filter_none({
        field.args['name']: field.to_cwl()
        for field in fields
      }).values(),
      {
        'id': 's',
        'inputBinding': {
          'prefix': '-s',
          'separate': True,
          'shellQuote': True,
        },
        'type': 'string?',
        'label': 'Location to send realtime update stream',
      },
      {
        'id': 'w',
        'inputBinding': {
          'prefix': '-w',
          'separate': True,
          'shellQuote': True,
        },
        'type': 'string',
        'default': 'output',
        'label': 'Work directory (all files are collected/output here)',
      },
    ],
    'outputs': [
      {
        'id': 'report',
        'type': 'Directory',
        'outputBinding': {
          'glob': '$(inputs.w)',
        },
      },
    ],
    's:programmingLanguage': 'Python',
    '$namespaces': {
      's': 'https://schema.org',
    },
    '$schemas': ['https://schema.org/version/latest/schemaorg-current-http.rdf'],
  })
  return schema


@nbinspect.command(help='Create CWL Tool definition for appyter')
@click.option('-o', '--output', default='-', type=click.File('w'), help='The output location of the inspection json')
@click.option('-i', '--info', default=None, type=click.File('r'), help='An appyter.json file for extra metadata as described in https://github.com/MaayanLab/appyter-catalog/blob/master/schema/appyter-validator.json')
@click_option_setenv('--cwd', envvar='APPYTER_CWD', default=os.getcwd(), help='The directory to treat as the current working directory for templates and execution')
@click_argument_setenv('ipynb', envvar='APPYTER_IPYNB')
def cwl(cwd, ipynb, info, output, **kwargs):
  cwd = os.path.realpath(cwd)
  env = get_jinja2_env(config=get_env(cwd=cwd, ipynb=ipynb, mode='inspect', **kwargs))
  with fsspec.open(join_slash(cwd, ipynb), 'r') as fr:
    nbtemplate = nb_from_ipynb_io(fr)
  info = json.load(info) if info is not None else None
  json.dump(render_cwl_from_nbtemplate(env, nbtemplate, cwd=cwd, ipynb=ipynb, info=info), output)



import contextlib
from appyter.ext.asyncio.helpers import ensure_sync

def assert_eq(a, b): assert a == b, f"{repr(a)} != {repr(b)}"

@contextlib.contextmanager
def assert_exc(ExpectedException):
  try:
    yield
    assert False, f"Expected {ExpectedException}, got no error"
  except ExpectedException:
    assert True

@contextlib.asynccontextmanager
async def _http_serve(directory, port=8888):
  ''' Serve a simple static listing of the given directory on the specified port
  '''
  from aiohttp import web
  from appyter.ext.aiohttp import with_app_running
  app = web.Application()
  app.add_routes([web.static('/subdir', str(directory), show_index=True)])
  async with with_app_running(app, port=port):
    yield

async def _http_connect(url, ssl_verify=False):
  ''' Successfully connect to and resolve a url
  '''
  import aiohttp
  async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=ssl_verify)) as session:
    async with session.get(url) as response:
      return await response.text()

@ensure_sync
@contextlib.asynccontextmanager
async def http_serve_ctx(directory, port=8888):
  ''' Context for _http_serve, ensuring server is ready before proceeding
  '''
  from appyter.ext.asyncio.try_n_times import async_try_n_times
  async with _http_serve(directory, port):
    await async_try_n_times(3, _http_connect, f"http://localhost:{port}")
    yield



import functools

def hashsum_io(algo, io, chunk_size=65536):
  import hashlib
  hashsum = hashlib.new(algo)
  while True:
    buf = io.read(chunk_size)
    if not buf: break
    hashsum.update(buf)
  return hashsum.hexdigest()

sha1sum_io = functools.partial(hashsum_io, 'sha1')

def hashsum_str(algo, obj):
  import hashlib
  sha1 = hashlib.new(algo)
  sha1.update(obj.encode())
  return sha1.hexdigest()

sha1sum_str = functools.partial(hashsum_str, 'sha1')

def hashsum_dict(algo, obj):
  import json
  return hashsum_str(algo, json.dumps(obj, sort_keys=True, separators=None, ensure_ascii=True))

sha1sum_dict = functools.partial(hashsum_dict, 'sha1')



import contextlib
import traceback
import logging

from appyter.ext.urllib import join_slash

logger = logging.getLogger(__name__)

from pathlib import Path, PurePosixPath
from fsspec import AbstractFileSystem, filesystem
from appyter.ext.fsspec.spec.mountable import MountableAbstractFileSystem
from appyter.ext.fsspec.spec.composable import ComposableAbstractFileSystem
from appyter.ext.pathlib.chroot import ChrootPurePosixPath

class ChrootFileSystem(MountableAbstractFileSystem, ComposableAbstractFileSystem, AbstractFileSystem):
  ''' chroot: update root and disallow access beyond chroot, only works on directories.
  '''
  root_marker = ''
  protocol = 'chroot'

  def __init__(self, target_protocol=None, target_options=None, fs=None, **kwargs):
    '''
    Parameters
    ----------
    target_protocol: str (optional)
      Target filesystem protocol. Provide either this or ``fs``.
    target_options: dict or None
      Passed to the instantiation of the FS, if fs is None.
    fs: filesystem instance
      The target filesystem to run against. Provide this or ``protocol``.
    '''
    super().__init__(**kwargs)
    if not (fs is None) ^ (target_protocol is None):
      raise ValueError(
          "Please provide one of filesystem instance (fs) or"
          " remote_protocol, not both"
      )
    self.kwargs = target_options or {}
    self.target_protocol = (
      target_protocol
      if isinstance(target_protocol, str)
      else (fs.protocol if isinstance(fs.protocol, str) else fs.protocol[0])
    )
    self.fs = fs if fs is not None else filesystem(target_protocol, **self.kwargs)

  def _target_fo(self):
    if '://' in self.storage_options.get('fo', ''):
      target_fo_split = self.storage_options.get('fo','').split('/')
      # [scheme://base-path][/path-part?...]
      return '/'.join(target_fo_split[:3]), '/' + '/'.join(target_fo_split[3:])
    else:
      return '', self.storage_options.get('fo', '')

  def _resolve_path(self, path):
    target_fo_base, target_fo_path = self._target_fo()
    resolved_path = join_slash(*filter(None, (target_fo_base, str((ChrootPurePosixPath(target_fo_path) / path).realpath()))))
    if resolved_path == '.':
      resolved_path = ''
    if resolved_path.startswith('./'):
      resolved_path = resolved_path[2:]
    return resolved_path

  def _unresolve_path(self, path):
    target_fo_base, target_fo_path = self._target_fo()
    if not path.startswith(target_fo_base):
      raise FileNotFoundError(path)
    path = path[len(target_fo_base):]
    # ensure leading/trailing slash is the same
    if path.startswith('/') and not target_fo_path.startswith('/'):
      path = path.lstrip('/')
    elif not path.startswith('/') and target_fo_path.startswith('/'):
      path = '/' + path
    unresolve_path = str(PurePosixPath(path).relative_to(target_fo_path))
    return unresolve_path

  @contextlib.contextmanager
  def __masquerade_os_error(self, path=None):
    try:
      yield
    except FileNotFoundError:
      if path:
        raise FileNotFoundError(path)
      else:
        raise FileNotFoundError
    except PermissionError:
      if path:
        raise PermissionError(path)
      else:
        raise PermissionError
    except OSError as e:
      logger.error(traceback.format_exc())
      if e.errno:
        raise type(e)(e.errno, f'{e.__class__.__name__} occurred, details have been hidden for security reasons')
      else:
        raise Exception(f"{e.__class__.__name__} occurred, details have been hidden for security reasons")
    except Exception as e:
      logger.error(traceback.format_exc())
      raise Exception(f"{e.__class__.__name__} occurred, details have been hidden for security reasons")

  def __enter__(self):
    if getattr(self.fs, '__enter__', None) is not None:
      with self.__masquerade_os_error():
        self.fs.__enter__()
    return self
  
  def __exit__(self, type, value, traceback):
    if getattr(self.fs, '__exit__', None) is not None:
      with self.__masquerade_os_error():
        self.fs.__exit__(type, value, traceback)

  @contextlib.contextmanager
  def mount(self, path='', mount_dir=None, fuse=True, passthrough=True, **kwargs):
    logger.debug(f"{self=} mount {mount_dir=} {fuse=}")
    with self.__masquerade_os_error(path=path):
      if passthrough and getattr(self.fs, 'mount', None) is not None:
        # try to use underlying fs mount
        with self.fs.mount(path=self._resolve_path(path), mount_dir=mount_dir, fuse=fuse, **kwargs) as mount_dir:
          yield mount_dir
      elif not fuse and 'file' in self.fs.protocol:
        # shortcut if we're talking about `file://` -- just use the actual directory
        yield Path(self._resolve_path(path))
      elif path:
        # just make another wrapper to deal with subpath mounting
        with ChrootFileSystem(fs=self.fs, fo=join_slash(self.storage_options.get('fo', ''), path)).mount(path='', mount_dir=mount_dir, fuse=fuse, passthrough=passthrough, **kwargs) as mount_dir:
          yield mount_dir
      elif fuse:
        # use fuse fs_mount
        from appyter.ext.asyncio.helpers import ensure_sync
        from appyter.ext.fsspec.fuse import fs_mount
        with ensure_sync(fs_mount(self, path, mount_dir=mount_dir)) as mount_dir:
          yield mount_dir
      else:
        # no fuse -- resort to copying files
        import shutil
        from appyter.ext.tempfile import tempdir
        # can't use fuse, default is to just copy files into the mount_dir
        with tempdir(mount_dir) as mount_dir:
          logger.debug(f"copying files over...")
          for f1_rel in self.glob('**', detail=True).values():
            f2_rel = mount_dir / f1_rel['name']
            f2_rel.parent.mkdir(parents=True, exist_ok=True)
            if not f2_rel.exists():
              if f1_rel['type'] == 'file':
                logger.debug(f"copying {f1_rel['name']} to {str(f2_rel)}")
                # TODO: if we're backed by a normal filesystem, make a link
                with self.open(f1_rel['name'], 'rb') as fr:
                  with f2_rel.open('wb') as fw:
                    shutil.copyfileobj(fr, fw)
              elif f1_rel['type'] == 'directory':
                logger.debug(f"making directory {f1_rel['name']}")
                f2_rel.mkdir(exist_ok=True)
              else:
                raise NotImplementedError
          logger.debug('Ready')
          try:
            yield mount_dir
          finally:
            logger.warning('Not copying back')

  def mkdir(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs.mkdir(self._resolve_path(path), **kwargs)

  def makedirs(self, path, exist_ok=False):
    with self.__masquerade_os_error(path=path):
      return self.fs.makedirs(self._resolve_path(path), exist_ok=exist_ok)

  def rmdir(self, path):
    with self.__masquerade_os_error(path=path):
      return self.fs.rmdir(self._resolve_path(path))

  def rm_file(self, path):
    with self.__masquerade_os_error(path=path):
      return self.fs.rm_file(self._resolve_path(path))

  def rm(self, path, recursive=False, maxdepth=None):
    with self.__masquerade_os_error(path=path):
      return self.fs.rm(self._resolve_path(path), recursive=recursive, maxdepth=maxdepth)

  def cat_file(self, path, start=None, end=None, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.cat_file(self._resolve_path(path), start=start, end=end, **kwargs)

  def put_file(self, lpath, rpath, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.put_file(lpath, self._resolve_path(rpath), **kwargs)

  def get_file(self, rpath, lpath, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.get_file(self._resolve_path(rpath), lpath, **kwargs)

  def cp_file(self, path1, path2, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.cp_file(self._resolve_path(path1), self._resolve_path(path2), **kwargs)

  def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.copy(self._resolve_path(path1), self._resolve_path(path2), recursive=recursive, on_error=on_error, **kwargs)

  def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
    with self.__masquerade_os_error():
      return self.fs.mv(self._resolve_path(path1), self._resolve_path(path2), recursive=recursive, maxdepth=maxdepth, **kwargs)

  def exists(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs.exists(self._resolve_path(path), **kwargs)

  def get_drs(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs.get_drs(self._resolve_path(path), **kwargs)

  def info(self, path, **kwargs):
    with self.__masquerade_os_error(path=path):
      info = self.fs.info(self._resolve_path(path), **kwargs)
      return dict(info, name=self._unresolve_path(info['name']))

  def ls(self, path, detail=True, **kwargs):
    logger.debug(f"ls({path}) => fs.ls({self._resolve_path(path)})")
    results = []
    with self.__masquerade_os_error(path=path):
      for f in self.fs.ls(self._resolve_path(path), detail=detail, **kwargs):
        if detail:
          f = dict(f, name=self._unresolve_path(f['name']))
          results.append(f)
        else:
          results.append(self._unresolve_path(f))
    return results

  def _open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
    with self.__masquerade_os_error(path=path):
      return self.fs._open(self._resolve_path(path), mode=mode, block_size=block_size, cache_options=cache_options, **kwargs)



import re
from appyter.fields import Field, FieldConstraintException
from appyter.render.flask_app.upload import upload_from_request
from appyter.ext.flask import secure_filepath
from appyter.ext.urllib import URI
from appyter.ext.re import re_full

class FileField(Field):
  r'''
  Representing a File URL.

  The field ends up being a string of the file path, which will be relative to the appyter notebook.

  .. code-block:: python

    import pandas as pd
    df = pd.read_csv({{ FileField(name='my-csv', label='My CSV') }})

  Importantly, a FileField ultimately resolves to a URL supporting several schemes. Appyters ensure
   that url is available as an actual file when the appyter executes.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param constraint: A regular expression for validating the file name.
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (str) A default value as an example and for use during prototyping
  :param storage: (Optional[str]) A storage prefix for browsing
  :param examples: (Optional[Dict[str, str]]) Named url paths to example files to upload
    paths can be relative i.e. `{ "my_file.txt": url_for('static', filename='my_file.txt') }`, or a remote url.
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, default=None, constraint=r'.*', valid_schemes=['drs', 's3', 'gs', 'ftp', 'http', 'https', 'storage', 'user'], **kwargs):
    super().__init__(
      constraint=constraint,
      default=default,
      valid_schemes=valid_schemes,
      **kwargs,
    )

  @property
  def raw_value(self):
    if type(self.args['value']) == str and self.args['value']:
      return self.args['value']
    else:
      return None

  def prepare_uri(self, raw_value):
    ''' The fully resolved URI for this file
    '''
    if raw_value is None: return None
    uri_parsed = URI(raw_value)
    if uri_parsed.scheme is None:
      if self._env.globals['_config']['SAFE_MODE']:
        uri_parsed = (
          URI(self._env.globals['_config']['PUBLIC_URL'])
            .join('static', uri_parsed.path)
            .with_query_string(uri_parsed.query_string)
            .with_fragment_path(secure_filepath(uri_parsed.fragment_name or uri_parsed.name))
            .with_fragment_query_string(uri_parsed.fragment_query_string)
        )
      else:
        uri_parsed = uri_parsed.with_scheme('file').with_fragment_path(raw_value)
    if uri_parsed.fragment_path is None:
      uri_parsed = uri_parsed.with_fragment_path(uri_parsed.name)
    return uri_parsed

  @property
  def uri(self):
    return self.prepare_uri(self.raw_value)

  def prepare(self, req):
    ''' Convert file in request into URI
    '''
    data_path = None
    # get data from request upload
    if getattr(req, 'files', None):
      data_path = upload_from_request(req, self.args['name'])
    # fallback to default (URI in request)
    if data_path is None: data_path = super().prepare(req)[self.args['name']]
    data_path = self.prepare_uri(data_path)
    data = {self.args['name']: str(data_path)}
    if data_path.fragment_path:
      data.update({ f"_file:{data_path.fragment_path}": str(data_path.with_fragment_path(None))})
    return data

  def constraint(self):
    if self.raw_value is None:
      return not self.args.get('required')
    else:
      if self._env.globals['_config']['SAFE_MODE']:
        if self.uri.scheme not in self.args['valid_schemes']:
          return False
      return re.match(re_full(self.args['constraint']), self.uri.fragment_path)

  @property
  def value(self):
    if self.uri is None:
      return None
    if not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    return self.uri.fragment_path

  def to_cwl(self):
    schema = super().to_cwl()
    schema['type'] = f"File{'' if self.args.get('required') == True else '?'}"
    return schema

  def to_cwl_value(self):
    return {
      'class': 'File',
      'path': str(self.uri.with_fragment_path(None)),
      'name': self.uri.fragment_path,
    } if self.raw_value is not None else None

  def to_jsonschema(self):
    schema = super().to_jsonschema()
    if self.args.get('constraint'): schema['pattern'] = re_full(self.args['constraint'])
    return schema

  @property
  def public_url(self):
    return str(self.uri.with_fragment_path(None))



import typing as t
import contextlib
import appyter.ext.fsspec
from appyter.ext.fsspec.core import url_to_fs_ex

import pytest
from appyter.ext.asyncio.helpers import ensure_sync
from appyter.ext.pytest import assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

@contextlib.asynccontextmanager
async def _drs_serve(port=8888):
  import re
  import aiohttp.web
  import datetime as dt
  from appyter import __version__
  from appyter.ext.aiohttp import with_app_running
  from appyter.ext.cryptography import adhoc_ssl_context
  app = aiohttp.web.Application()
  routes = aiohttp.web.RouteTableDef()
  @routes.get('/ga4gh/drs/v1/service-info')
  async def drs_service_info(request: aiohttp.web.Request):
    return aiohttp.web.json_response({
      'id': f"cloud.maayanlab.appyter",
      'name': 'Appyter',
      'type': {
        'group': 'cloud.maayanlab.appyter',
        'artifact': 'drs',
        'version': '1.0.0',
      },
      'organization': {
        'name': 'MaayanLab',
        'url': 'https://maayanlab.cloud',
      },
      'version': __version__,
    })
  @routes.get('/ga4gh/drs/v1/objects/{file_id}')
  async def drs_object(request: aiohttp.web.Request):
    file_id = request.match_info['file_id']
    if file_id == '12345':
      return aiohttp.web.json_response({
        'id': file_id,
        'self_uri': f"{re.sub(r'^https?://', 'drs://', str(request.url.origin()))}/{file_id}",
        'access_methods': [
          {
            'access_id': 'primary',
            'type': 'https',
          },
        ],
        'checksums': [
          {
            'type': 'sha-1',
            'checksum': file_id,
          },
        ],
        'created_time': dt.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ'),
        'size': 1,
      })
    else:
      raise aiohttp.web.HTTPNotFound()
  @routes.get('/ga4gh/drs/v1/objects/{file_id}/access/{access_id}')
  async def drs_object_access(request: aiohttp.web.Request):
    file_id = request.match_info['file_id']
    access_id = request.match_info['access_id']
    if access_id == 'primary' and file_id == '12345':
      return aiohttp.web.json_response({ 'url': f"{str(request.url.origin())}/ga4gh/drs/v1/objects/{file_id}/fetch" })
    else:
      raise aiohttp.web.HTTPNotFound()
  @routes.get('/ga4gh/drs/v1/objects/{file_id}/fetch')
  async def drs_object_fetch(request: aiohttp.web.Request):
    file_id = request.match_info['file_id']
    if file_id == '12345':
      return aiohttp.web.Response(body=b'A')
    else:
      raise aiohttp.web.HTTPNotFound()
  app.add_routes(routes)
  async with adhoc_ssl_context(cn=f"localhost:{port}") as ssl_context:
    async with with_app_running(app, port=port, ssl_context=ssl_context):
      yield

@ensure_sync
@contextlib.asynccontextmanager
async def drs_serve_ctx(port=8888):
  ''' Context for _drs_serve, ensuring server is ready before proceeding
  '''
  from appyter.ext.pytest import _http_connect
  from appyter.ext.asyncio.try_n_times import async_try_n_times
  async with _drs_serve(port):
    await async_try_n_times(3, _http_connect, f"https://localhost:{port}/ga4gh/drs/v1/service-info")
    yield

def test_drs():
  ''' Serve the test directory over http and ensure chroot_fs still works using the webserver
  '''
  port = 8888
  with drs_serve_ctx(port):
    fs, fo = url_to_fs_ex(f"drs://localhost:{port}/12345", ssl_verify=False)
    with fs as fs:
      with fs.open(fo, 'rb') as fr:
        assert_eq(fr.read(), b'A')



import asyncio
import aiohttp, aiohttp.web
import contextlib
import logging
from appyter.ext.asyncio.helpers import ensure_sync

logger = logging.getLogger(__name__)

@contextlib.asynccontextmanager
async def with_app_running(app, host='127.0.0.1', port=5000, path=None, ssl_context=None):
  runner = aiohttp.web.AppRunner(app)
  await runner.setup()
  if path:
    site = aiohttp.web.UnixSite(runner, path, ssl_context=ssl_context)
    logger.info(f"Starting server at unix://{path}")
  elif port is not None:
    site = aiohttp.web.TCPSite(runner, host, port, ssl_context=ssl_context)
    logger.info(f"Starting server on http://{host}:{port}")
  else:
    raise NotImplementedError
  await site.start()
  try:
    yield
  finally:
    logger.info(f"Cleaning up server...")
    await runner.cleanup()

@ensure_sync
async def run_app(app, host='127.0.0.1', port=5000, path=None):
  async with with_app_running(app, host=host, port=port, path=path):
    try:
      await asyncio.Event().wait()
    except asyncio.CancelledError:
      pass



import contextlib

@contextlib.asynccontextmanager
async def ensure_storage(storage_uri):
  import fsspec
  from appyter.ext.fsspec.core import url_to_fs_ex
  if 'storage' not in fsspec.registry:
    from appyter.ext.fsspec.singleton import SingletonFileSystem
    from appyter.ext.asyncio.helpers import ensure_async_contextmanager
    fs, fo = url_to_fs_ex(storage_uri)
    async with ensure_async_contextmanager(SingletonFileSystem(proto='storage', fs=fs, fo=fo)) as fs:
      yield fs
  else:
    fs, _ = url_to_fs_ex('storage://')
    yield fs



from appyter.fields import Field, FieldConstraintException
from appyter.ext.json import try_json_loads

class MultiChoiceField(Field):
  r''' Represing a multi-selectable combo box.
  The resulting selection is represented with a list of strings that were chosen.

  :param name: (str) A name that will be used to refer to the object as a variable and in the HTML form.
  :param label: (str) A human readable label for the field for the HTML form
  :param description: (Optional[str]) A long human readable description for the field for the HTML form
  :param choices: (Union[List[str], Dict[str, str]]) A set of choices that are available for this field or lookup table mapping from choice label to resulting value
  :param required: (Optional[bool]) Whether or not this field is required (defaults to false)
  :param default: (float) A default value as an example and for use during prototyping
  :param section: (Optional[str]) The name of a SectionField for which to nest this field under, defaults to a root SectionField
  :param value: (INTERNAL Any) The raw value of the field (from the form for instance)
  :param \**kwargs: Remaining arguments passed down to :class:`appyter.fields.Field`'s constructor.
  '''
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    assert self.choices, 'MultiChoiceField requires choices'

  @property
  def raw_value(self):
    value = try_json_loads(self.args['value'])
    if not value:
      return []
    elif type(value) == list:
      return value
    elif type(value) == str:
      return [value]
    else:
      return [self.args['value']]

  def constraint(self):
    if not self.raw_value:
      return not self.args.get('required')
    else:
      return type(self.raw_value) == list and all(v in self.choices for v in self.raw_value)

  @property
  def value(self):
    if not self.constraint():
      raise FieldConstraintException(
        field=self.field,
        field_name=self.args['name'],
        value=self.raw_value,
      )
    elif type(self.choices) == dict:
      return [self.choices[v] for v in self.raw_value]
    else:
      return self.raw_value

  def to_jsonschema(self):
    schema = {'type': 'array', 'items': { 'type': 'string' } }
    if self.args.get('label'): schema['title'] = self.args['label']
    if self.args.get('description'): schema['description'] = self.args['description']
    if self.args.get('choices'): schema['items']['enum'] = list(self.args['choices'])
    if self.args.get('default'): schema['default'] = self.args['default']
    return schema

  def to_cwl(self):
    schema = super().to_cwl()
    # NOTE: CWL's array enum is broken upstream
    # if self.args.get('required') == True:
    #   schema['type'] = {
    #     'type': 'array',
    #     'items': {
    #       'type': 'enum',
    #       'symbols': list(self.args['choices'])
    #     },
    #   }
    # else:
    #   schema['type'] = ['null', {
    #     'type': 'array',
    #     'items': {
    #       'type': 'enum',
    #       'symbols': list(self.args['choices'])
    #     },
    #   }]
    schema['type'] = f"string{'' if self.args.get('required') == True else '?'}"
    return schema

  def to_cwl_value(self):
    import json
    return json.dumps(self.raw_value)

  def to_click(self):
    args, kwargs = super().to_click()
    # kwargs['multiple'] = True
    import click
    kwargs['type'] = click.STRING
    return args, kwargs



import appyter.ext.multiprocessing

import tempfile
import contextlib
from pathlib import Path

import appyter.ext.fsspec
from appyter.ext.fsspec.core import url_to_chroot_fs

import pytest
from appyter.ext.pytest import http_serve_ctx, assert_eq
from appyter.ext.asyncio.event_loop import with_event_loop
@pytest.fixture(scope="session", autouse=True)
def event_loop_fixture():
  with with_event_loop():
    yield

@contextlib.contextmanager
def _test_ctx():
  with tempfile.TemporaryDirectory() as tmpdir:
    tmpdir = Path(tmpdir)
    (tmpdir/'a'/'b').mkdir(parents=True)
    with (tmpdir/'a'/'b'/'c').open('w') as fw:
      fw.write('C')
    with (tmpdir/'a'/'d').open('w') as fw:
      fw.write('D')
    with (tmpdir/'e').open('w') as fw:
      fw.write('E')
    yield tmpdir

def test_file_chroot_in_chroot():
  with _test_ctx() as tmpdir:
    # instantiate filesystem & check that it works
    with url_to_chroot_fs(f"chroot::file://{str(tmpdir)}") as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('a/d'), b'D')
      assert_eq(fs.cat('e'), b'E')

def test_file_chroot():
  with _test_ctx() as tmpdir:
    # instantiate filesystem & check that it works
    with url_to_chroot_fs(str(tmpdir)) as fs:
      assert_eq(frozenset(fs.glob('**')), frozenset(['a', 'a/b', 'a/b/c', 'a/d', 'e']))
      assert_eq(fs.cat('a/b/c'), b'C')
      assert_eq(fs.cat('a/d'), b'D')
      assert_eq(fs.cat('e'), b'E')

def test_http_chroot():
  ''' Serve the test directory over http and ensure chroot_fs still works using the webserver
  '''
  port = 8888
  with _test_ctx() as tmpdir:
    with http_serve_ctx(tmpdir, port):
      # instantiate filesystem & check that it works
      with url_to_chroot_fs(f"http://localhost:{port}/subdir/") as fs:
        assert_eq(fs.cat('a/b/c'), b'C')
        assert_eq(fs.cat('a/d'), b'D')
        assert_eq(fs.cat('e'), b'E')



import re
import itertools
import dataclasses
import typing as t
import urllib.parse
from appyter.ext.pathlib.chroot import ChrootPurePosixPath
from appyter.ext.dict import dict_merge, expand_dotmap

def parent_url(url):
  parent, *filename = url.rsplit('/', maxsplit=1)
  return parent if filename else ''

def url_filename(url):
  parent, *filename = url.rsplit('/', maxsplit=1)
  return filename[0] if filename else parent

def join_slash(*parts):
  if not parts: return ''
  part0, *parts = parts
  return '/'.join(itertools.chain((part0.rstrip('/'),), (part.lstrip('/') for part in parts)))

def join_url(root, *parts):
  return join_slash(root, str(ChrootPurePosixPath('/').join(*parts)))

url_expr = re.compile(
  r'^((?P<scheme>.+?)://(?P<authority>((?P<username>[^/:@\?#]+?)(:(?P<password>[^/@\?#]+?))?@)?(?P<netloc>(?P<hostname>[^:/\?#]+)(:(?P<port>\d+))?))?)?(?P<path>.*?)(\?(?P<query_string>.*?))?(#(?P<fragment>.*?))?$'
)
fragment_expr =  re.compile(
  r'^(?P<path>.*?)(\?(?P<query_string>.*?))?$'
)

def parse_qs_values(query_map):
  from appyter.ext.itertools import collapse
  from appyter.ext.json import try_json_loads
  return {
    k: collapse([try_json_loads(v) if v != '' else True for v in V])
    for k, V in query_map.items()
  }

@dataclasses.dataclass(init=False, repr=False, frozen=True)
class URI:
  ''' Not unlike yarl's URL class but
  - support for `::` notation as used in fsspec URIs
  - posix_path path operation
  - fragment parsing
  - dotmap support (query_ex)
  '''
  scheme: t.Optional[str]
  username: t.Optional[str]
  password: t.Optional[str]
  hostname: t.Optional[str]
  port: t.Optional[int]
  path: str
  query_string: t.Optional[str]
  fragment: t.Optional[str]

  def __init__(self, url=None, scheme=None, username=None, password=None, hostname=None, port=None, path=None, query_string=None, fragment=None):
    if url is not None:
      m = url_expr.match(url)
      object.__setattr__(self, 'scheme', m.group('scheme'))
      object.__setattr__(self, 'username', m.group('username'))
      object.__setattr__(self, 'password', m.group('password'))
      object.__setattr__(self, 'hostname', m.group('hostname'))
      object.__setattr__(self, 'port', int(m.group('port')) if m.group('port') is not None else None)
      object.__setattr__(self, 'path', m.group('path'))
      object.__setattr__(self, 'query_string', m.group('query_string'))
      object.__setattr__(self, 'fragment', m.group('fragment'))
    else:
      object.__setattr__(self, 'scheme', scheme)
      object.__setattr__(self, 'username', username)
      object.__setattr__(self, 'password', password)
      object.__setattr__(self, 'hostname', hostname)
      object.__setattr__(self, 'port', int(port) if port is not None else None)
      object.__setattr__(self, 'path', path or '')
      object.__setattr__(self, 'query_string', query_string)
      object.__setattr__(self, 'fragment', fragment)

  @property
  def netloc(self):
    if self.hostname is not None and self.port is not None:
      return f"{self.hostname}:{self.port}"
    elif self.hostname is not None:
      return self.hostname
    else:
      return None

  @property
  def auth(self):
    if self.username is not None:
      if self.password is not None:
        return f"{self.username}:{self.password}"
      else:
        return self.username
    return None

  @property
  def authority(self):
    if self.netloc is not None:
      if self.auth is not None:
        return f"{self.auth}@{self.netloc}"
      else:
        return self.netloc
    return None

  @property
  def posix_path(self):
    return ChrootPurePosixPath(self.path)

  @property
  def name(self):
    return self.posix_path.root.name

  @property
  def parent(self):
    return self.with_path(str(self.posix_path.root.parent))

  @property
  def query(self):
    return urllib.parse.parse_qs(self.query_string) if self.query_string is not None else None

  @property
  def query_ex(self):
    return expand_dotmap(parse_qs_values(self.query)) if self.query is not None else {}

  @property
  def fragment_path(self):
    return fragment_expr.match(self.fragment).group('path') if self.fragment is not None else None

  @property
  def fragment_posix_path(self):
    return ChrootPurePosixPath(self.fragment_path) if self.fragment_path is not None else None

  @property
  def fragment_name(self):
    return self.fragment_posix_path.root.name if self.fragment_posix_path is not None else None

  @property
  def fragment_parent(self):
    return self.with_path(str(self.fragment_posix_path.root.parent)) if self.fragment_posix_path is not None else None

  @property
  def fragment_query_string(self):
    return fragment_expr.match(self.fragment).group('query_string') if self.fragment is not None else None

  @property
  def fragment_query(self):
    return urllib.parse.parse_qs(self.fragment_query_string) if self.fragment_query_string is not None else None

  @property
  def fragment_query_ex(self):
    return expand_dotmap(parse_qs_values(self.fragment_query)) if self.fragment_query is not None else {}

  def __str__(self):
    return ''.join(filter(None, (
      f"{self.scheme}://" if self.scheme is not None else None,
      join_url(self.authority, self.path) if self.authority is not None else self.path,
      f"?{self.query_string}" if self.query_string is not None else None,
      f"#{self.fragment}" if self.fragment is not None else None,
    )))

  def __repr__(self):
    return "{}('{}')".format(self.__class__.__name__, str(self))

  def with_scheme(self, scheme):
    return URI(
      scheme=scheme,
      username=self.username,
      password=self.password,
      hostname=self.hostname,
      port=self.port,
      path=self.path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_username(self, username):
    return URI(
      scheme=self.scheme,
      username=username,
      password=self.password,
      hostname=self.hostname,
      port=self.port,
      path=self.path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_password(self, password):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=password,
      hostname=self.hostname,
      port=self.port,
      path=self.path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_hostname(self, hostname):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=self.password,
      hostname=hostname,
      port=self.port,
      path=self.path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_port(self, port):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=self.password,
      hostname=self.hostname,
      port=port,
      path=self.path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_path(self, path):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=self.password,
      hostname=self.hostname,
      port=self.port,
      path=path,
      query_string=self.query_string,
      fragment=self.fragment,
    )

  def with_query_string(self, query_string):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=self.password,
      hostname=self.hostname,
      port=self.port,
      path=self.path,
      query_string=query_string,
      fragment=self.fragment,
    )

  def with_query(self, query):
    return self.with_query_string(urllib.parse.urlencode(query, doseq=True) if query is not None else None)

  def update_query(self, query):
    return self.with_query(dict_merge(self.query or {}, **query))

  def with_fragment(self, fragment):
    return URI(
      scheme=self.scheme,
      username=self.username,
      password=self.password,
      hostname=self.hostname,
      port=self.port,
      path=self.path,
      query_string=self.query_string,
      fragment=fragment,
    )

  def with_fragment_path(self, fragment_path):
    if fragment_path is None and self.fragment_query_string is None:
      return self.with_fragment(None)
    elif self.fragment_query_string is None:
      return self.with_fragment(fragment_path)
    else:
      return self.with_fragment(f"{fragment_path or ''}?{self.fragment_query_string}")

  def with_fragment_query_string(self, fragment_query_string):
    if self.fragment_path is None and fragment_query_string is None:
      return self.with_fragment(None)
    elif fragment_query_string is None:
      return self.with_fragment(self.fragment_path)
    else:
      return self.with_fragment(f"{self.fragment_path or ''}?{fragment_query_string}")

  def with_fragment_query(self, fragment_query):
    return self.with_fragment_query_string(urllib.parse.urlencode(fragment_query, doseq=True))

  def update_fragment_query(self, fragment_query):
    return self.with_fragment_query(dict_merge(self.fragment_query or {}, **fragment_query))

  def join(self, *parts):
    return self.with_path(str(self.posix_path.join(*parts).realpath())) if self.posix_path is not None else None

  def fragment_join(self, *parts):
    return self.with_fragment_path(str(self.fragment_posix_path.join(*parts).realpath())) if self.fragment_posix_path is not None else None



import asyncio
import typing as t
from contextlib import asynccontextmanager
from nbclient import NotebookClient
from nbformat import NotebookNode
from nbclient.exceptions import CellExecutionComplete, DeadKernelError, CellControlSignal
from nbclient.util import run_hook
from nbclient.client import timestamp
from appyter.ext.asyncio.event_loop import get_event_loop
from appyter.ext.asyncio.helpers import ensure_async

class NotebookClientIOPubHook(NotebookClient):
  ''' A notebook client with the ability to hook into iopub updates
  '''
  def __init__(self, *args, iopub_hook=None, **kwargs):
    super().__init__(*args, **kwargs)
    self.iopub_hook = iopub_hook

  async def _async_poll_output_msg(
    self,
    parent_msg_id,
    cell,
    cell_index
  ):
    ''' Override output_msg polling to send to iopub_hook
    '''
    assert self.kc is not None
    complete = False
    while not complete:
      msg = await ensure_async(self.kc.iopub_channel.get_msg(timeout=None))
      if msg['parent_header'].get('msg_id') == parent_msg_id:
        try:
          # Will raise CellExecutionComplete when completed
          await self.process_message(msg, cell, cell_index)
        except CellExecutionComplete:
          complete = True
        finally:
          if self.iopub_hook is not None:
            await self.iopub_hook(cell, cell_index)

  async def handle_comm_msg(
    self,
    outs: t.List,
    msg: t.Dict,
    cell_index: int
  ) -> None:
    ''' Store ipywidget state in each cell during execution and trigger iopub_hook
    '''
    super().handle_comm_msg(outs, msg, cell_index)
    content = msg['content']
    data = content['data']
    if self.store_widget_state and 'state' in data:  # ignore custom msg'es
      # Here we store the updated state in the cell during execution, and trigger an iopub_hook
      # we'll arbitrarily choose to store it in the cell's metadata.widget-state
      cell = self.nb['cells'][cell_index]
      cell.setdefault('metadata', {}).setdefault('widgets', {
        'version_major': 2,
        'version_minor': 0,
      }).setdefault('state', {}).update({
        content['comm_id']: self._serialize_widget_state(self.widget_state[content['comm_id']]),
      })
      for key, widget in cell['metadata']['widgets']['state'].items():
        buffers = self.widget_buffers.get(key)
        if buffers:
            widget['buffers'] = buffers
      if self.iopub_hook is not None:
        await self.iopub_hook(cell, cell_index)

  def _kc_execute(self, *args, **kwargs):
    return self.kc.execute(*args, **kwargs)

  async def async_execute_cell(
      self,
      cell: NotebookNode,
      cell_index: int,
      execution_count: t.Optional[int] = None,
      store_history: bool = True) -> NotebookNode:
    """
    Executes a single code cell.

    To execute all cells see :meth:`execute`.

    Parameters
    ----------
    cell : nbformat.NotebookNode
      The cell which is currently being processed.
    cell_index : int
      The position of the cell within the notebook object.
    execution_count : int
      The execution count to be assigned to the cell (default: Use kernel response)
    store_history : bool
      Determines if history should be stored in the kernel (default: False).
      Specific to ipython kernels, which can store command histories.

    Returns
    -------
    output : dict
      The execution output payload (or None for no output).

    Raises
    ------
    CellExecutionError
      If execution failed and should raise an exception, this will be raised
      with defaults about the failure.

    Returns
    -------
    cell : NotebookNode
      The cell which was just processed.
    """
    assert self.kc is not None
    if cell.cell_type != 'code' or not cell.source.strip():
      self.log.debug("Skipping non-executing cell %s", cell_index)
      return cell

    if self.record_timing and 'execution' not in cell['metadata']:
      cell['metadata']['execution'] = {}

    self.log.debug("Executing cell:\n%s", cell.source)
    parent_msg_id = await ensure_async(self._kc_execute)(
      cell.source,
      store_history=store_history,
      stop_on_error=not self.allow_errors
    )
    # We launched a code cell to execute
    self.code_cells_executed += 1
    exec_timeout = self._get_timeout(cell)

    cell.outputs = []
    self.clear_before_next_output = False

    task_poll_kernel_alive = asyncio.ensure_future(
      self._async_poll_kernel_alive()
    )
    task_poll_output_msg = asyncio.ensure_future(
      self._async_poll_output_msg(parent_msg_id, cell, cell_index)
    )
    self.task_poll_for_reply = asyncio.ensure_future(
      self._async_poll_for_reply(
        parent_msg_id, cell, exec_timeout, task_poll_output_msg, task_poll_kernel_alive
      )
    )
    try:
      exec_reply = await self.task_poll_for_reply
    except asyncio.CancelledError:
      # can only be cancelled by task_poll_kernel_alive when the kernel is dead
      task_poll_output_msg.cancel()
      raise DeadKernelError("Kernel died")
    except Exception as e:
      # Best effort to cancel request if it hasn't been resolved
      try:
        # Check if the task_poll_output is doing the raising for us
        if not isinstance(e, CellControlSignal):
          task_poll_output_msg.cancel()
      finally:
        raise

    if execution_count:
      cell['execution_count'] = execution_count
    await self._check_raise_for_error(cell, cell_index, exec_reply)
    self.nb['cells'][cell_index] = cell
    return cell

  @asynccontextmanager
  async def async_setup_kernel(self, **kwargs) -> t.AsyncGenerator:
      """
      Context manager for setting up the kernel to execute a notebook.

      This assigns the Kernel Manager (``self.km``) if missing and Kernel Client(``self.kc``).

      When control returns from the yield it stops the client's zmq channels, and shuts
      down the kernel.
      """
      # by default, cleanup the kernel client if we own the kernel manager
      # and keep it alive if we don't
      cleanup_kc = kwargs.pop('cleanup_kc', self.owns_km)
      if self.km is None:
          self.km = self.create_kernel_manager()

      loop = get_event_loop()

      if not self.km.has_kernel:
          await self.async_start_new_kernel(**kwargs)
          await self.async_start_new_kernel_client()
      try:
          yield
      except RuntimeError as e:
          await run_hook(self.on_notebook_error, notebook=self.nb)
          raise e
      finally:
          if cleanup_kc:
              await self._async_cleanup_kernel()
          await run_hook(self.on_notebook_complete, notebook=self.nb)

  async def process_message(
      self,
      msg: t.Dict,
      cell: NotebookNode,
      cell_index: int) -> t.Optional[t.List]:
    """
    Processes a kernel message, updates cell state, and returns the
    resulting output object that was appended to cell.outputs.

    The input argument *cell* is modified in-place.

    Parameters
    ----------
    msg : dict
      The kernel message being processed.
    cell : nbformat.NotebookNode
      The cell which is currently being processed.
    cell_index : int
      The position of the cell within the notebook object.

    Returns
    -------
    output : dict
      The execution output payload (or None for no output).

    Raises
    ------
    CellExecutionComplete
      Once a message arrives which indicates computation completeness.

    """
    msg_type = msg['msg_type']
    self.log.debug("msg_type: %s", msg_type)
    content = msg['content']
    self.log.debug("content: %s", content)

    display_id = content.get('transient', {}).get('display_id', None)
    if display_id and msg_type in {'execute_result', 'display_data', 'update_display_data'}:
      self._update_display_id(display_id, msg)

    # set the prompt number for the input and the output
    if 'execution_count' in content:
      cell['execution_count'] = content['execution_count']

    if self.record_timing:
      if msg_type == 'status':
        if content['execution_state'] == 'idle':
          cell['metadata']['execution']['iopub.status.idle'] = timestamp()
        elif content['execution_state'] == 'busy':
          cell['metadata']['execution']['iopub.status.busy'] = timestamp()
      elif msg_type == 'execute_input':
        cell['metadata']['execution']['iopub.execute_input'] = timestamp()

    if msg_type == 'status':
      if content['execution_state'] == 'idle':
        raise CellExecutionComplete()
    elif msg_type == 'clear_output':
      self.clear_output(cell.outputs, msg, cell_index)
    elif msg_type.startswith('comm'):
      await self.handle_comm_msg(cell.outputs, msg, cell_index)
    # Check for remaining messages we don't process
    elif msg_type not in ['execute_input', 'update_display_data']:
      # Assign output as our processed "result"
      return self.output(cell.outputs, msg, display_id, cell_index)
    return None
